#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The Filter test suite.
"""
import gzip
import warnings

import numpy as np
import scipy.signal as sg

from obspy import read
from obspy.signal.filter import (bandpass, bandstop, highpass, lowpass,
                                 envelope, lowpass_cheby_2)


class TestFilter():
    """
    Test cases for Filter.
    """

    def test_bandpass_vs_pitsa(self, testdata):
        """
        Test Butterworth bandpass filter against Butterworth bandpass filter
        of PITSA. Note that the corners value is twice the value of the filter
        sections in PITSA. The rms of the difference between ObsPy and PITSA
        tends to get bigger with higher order filtering.
        """
        # load test file
        with gzip.open(testdata['rjob_20051006.gz']) as f:
            data = np.loadtxt(f)
        # parameters for the test
        samp_rate = 200.0
        freq1 = 5
        freq2 = 10
        corners = 4
        # filter trace
        datcorr = bandpass(data, freq1, freq2, df=samp_rate, corners=corners)
        # load pitsa file
        with gzip.open(testdata['rjob_20051006_bandpass.gz']) as f:
            data_pitsa = np.loadtxt(f)
        # calculate normalized rms
        rms = np.sqrt(np.sum((datcorr - data_pitsa) ** 2) /
                      np.sum(data_pitsa ** 2))
        assert rms < 1.0e-05

    def test_bandpass_zphsh_vs_pitsa(self, testdata):
        """
        Test Butterworth zero-phase bandpass filter against Butterworth
        zero-phase bandpass filter of PITSA. Note that the corners value is
        twice the value of the filter sections in PITSA. The rms of the
        difference between ObsPy and PITSA tends to get bigger with higher
        order filtering.
        Note: The Zero-Phase filters deviate from PITSA's zero-phase filters
        at the end of the trace! The rms for the test is calculated omitting
        the last 200 samples, as this part of the trace is assumed to
        generally be of low interest/importance.
        """
        # load test file
        with gzip.open(testdata['rjob_20051006.gz']) as f:
            data = np.loadtxt(f)
        # parameters for the test
        samp_rate = 200.0
        freq1 = 5
        freq2 = 10
        corners = 2
        # filter trace
        datcorr = bandpass(data, freq1, freq2, df=samp_rate,
                           corners=corners, zerophase=True)
        # load pitsa file
        with gzip.open(testdata['rjob_20051006_bandpassZPHSH.gz']) as f:
            data_pitsa = np.loadtxt(f)
        # calculate normalized rms
        rms = np.sqrt(np.sum((datcorr[:-200] - data_pitsa[:-200]) ** 2) /
                      np.sum(data_pitsa[:-200] ** 2))
        assert rms < 1.0e-05

    def test_lowpass_vs_pitsa(self, testdata):
        """
        Test Butterworth lowpass filter against Butterworth lowpass filter of
        PITSA. Note that the corners value is twice the value of the filter
        sections in PITSA. The rms of the difference between ObsPy and PITSA
        tends to get bigger with higher order filtering.
        """
        # load test file
        with gzip.open(testdata['rjob_20051006.gz']) as f:
            data = np.loadtxt(f)
        # parameters for the test
        samp_rate = 200.0
        freq = 5
        corners = 4
        # filter trace
        datcorr = lowpass(data, freq, df=samp_rate, corners=corners)
        # load pitsa file
        with gzip.open(testdata['rjob_20051006_lowpass.gz']) as f:
            data_pitsa = np.loadtxt(f)
        # calculate normalized rms
        rms = np.sqrt(np.sum((datcorr - data_pitsa) ** 2) /
                      np.sum(data_pitsa ** 2))
        assert rms < 1.0e-05

    def test_lowpass_zphsh_vs_pitsa(self, testdata):
        """
        Test Butterworth zero-phase lowpass filter against Butterworth
        zero-phase lowpass filter of PITSA. Note that the corners value is
        twice the value of the filter sections in PITSA. The rms of the
        difference between ObsPy and PITSA tends to get bigger with higher
        order filtering.
        Note: The Zero-Phase filters deviate from PITSA's zero-phase filters
        at the end of the trace! The rms for the test is calculated omitting
        the last 200 samples, as this part of the trace is assumed to
        generally be of low interest/importance.
        """
        # load test file
        filename = testdata['rjob_20051006.gz']
        with gzip.open(filename) as f:
            data = np.loadtxt(f)
        # parameters for the test
        samp_rate = 200.0
        freq = 5
        corners = 2
        # filter trace
        datcorr = lowpass(data, freq, df=samp_rate, corners=corners,
                          zerophase=True)
        # load pitsa file
        filename = testdata['rjob_20051006_lowpassZPHSH.gz']
        with gzip.open(filename) as f:
            data_pitsa = np.loadtxt(f)
        # calculate normalized rms
        rms = np.sqrt(np.sum((datcorr[:-200] - data_pitsa[:-200]) ** 2) /
                      np.sum(data_pitsa[:-200] ** 2))
        assert rms < 1.0e-05

    def test_highpass_vs_pitsa(self, testdata):
        """
        Test Butterworth highpass filter against Butterworth highpass filter
        of PITSA. Note that the corners value is twice the value of the filter
        sections in PITSA. The rms of the difference between ObsPy and PITSA
        tends to get bigger with higher order filtering.
        """
        # load test file
        filename = testdata['rjob_20051006.gz']
        with gzip.open(filename) as f:
            data = np.loadtxt(f)
        # parameters for the test
        samp_rate = 200.0
        freq = 10
        corners = 4
        # filter trace
        datcorr = highpass(data, freq, df=samp_rate, corners=corners)
        # load pitsa file
        filename = testdata['rjob_20051006_highpass.gz']
        with gzip.open(filename) as f:
            data_pitsa = np.loadtxt(f)
        # calculate normalized rms
        rms = np.sqrt(np.sum((datcorr - data_pitsa) ** 2) /
                      np.sum(data_pitsa ** 2))
        assert rms < 1.0e-05

    def test_highpass_zphsh_vs_pitsa(self, testdata):
        """
        Test Butterworth zero-phase highpass filter against Butterworth
        zero-phase highpass filter of PITSA. Note that the corners value is
        twice the value of the filter sections in PITSA. The rms of the
        difference between ObsPy and PITSA tends to get bigger with higher
        order filtering.
        Note: The Zero-Phase filters deviate from PITSA's zero-phase filters
        at the end of the trace! The rms for the test is calculated omitting
        the last 200 samples, as this part of the trace is assumed to
        generally be of low interest/importance.
        """
        # load test file
        filename = testdata['rjob_20051006.gz']
        with gzip.open(filename) as f:
            data = np.loadtxt(f)
        # parameters for the test
        samp_rate = 200.0
        freq = 10
        corners = 2
        # filter trace
        datcorr = highpass(data, freq, df=samp_rate, corners=corners,
                           zerophase=True)
        # load pitsa file
        filename = testdata['rjob_20051006_highpassZPHSH.gz']
        with gzip.open(filename) as f:
            data_pitsa = np.loadtxt(f)
        # calculate normalized rms
        rms = np.sqrt(np.sum((datcorr[:-200] - data_pitsa[:-200]) ** 2) /
                      np.sum(data_pitsa[:-200] ** 2))
        assert rms < 1.0e-05

    def test_envelope_vs_pitsa(self, testdata):
        """
        Test Envelope filter against PITSA.
        The rms is not so good, but the fit is still good in most parts.
        """
        # load test file
        filename = testdata['rjob_20051006.gz']
        with gzip.open(filename) as f:
            data = np.loadtxt(f)
        # filter trace
        datcorr = envelope(data)
        # load pitsa file
        filename = testdata['rjob_20051006_envelope.gz']
        with gzip.open(filename) as f:
            data_pitsa = np.loadtxt(f)
        # calculate normalized rms
        rms = np.sqrt(np.sum((datcorr - data_pitsa) ** 2) /
                      np.sum(data_pitsa ** 2))
        assert rms < 1.0e-02

    def test_lowpass_cheby_2(self):
        """
        Check magnitudes of basic lowpass cheby2
        """
        df = 200  # Hz
        b, a = lowpass_cheby_2(data=None, freq=50,
                               df=df, maxorder=12, ba=True)
        nyquist = 100
        # calculate frequency response
        w, h = sg.freqz(b, a, nyquist)
        freq = w / np.pi * nyquist
        h_db = 20 * np.log10(abs(h))
        # be smaller than -96dB above lowpass frequency
        assert -96 > h_db[freq > 50].max()
        # be 0 (1dB ripple) before filter ramp
        assert h_db[freq < 25].min() > -1

    def test_bandpass_high_corner_at_nyquist(self):
        """
        Check that using exactly Nyquist for high corner gives correct results.
        See #1451.
        """
        tr = read()[0]
        data = tr.data[:1000]

        df = tr.stats.sampling_rate
        nyquist = df / 2.0

        for low_corner in (6.0, 8.55, 8.59):
            for corners in (3, 4, 5, 6):
                # this is filtering with high corner slightly below what we
                # catch and change into highpass
                with warnings.catch_warnings(record=True) as w:
                    warnings.simplefilter("always")
                    expected = bandpass(
                        data, low_corner, nyquist * (1 - 1.1e-6), df=df,
                        corners=corners)
                    assert len(w) == 0
                # all of these should be changed into a highpass
                with warnings.catch_warnings(record=True) as w:
                    warnings.simplefilter("always")
                    got1 = bandpass(data, low_corner, nyquist * (1 - 0.9e-6),
                                    df=df, corners=corners)
                    got2 = bandpass(data, low_corner, nyquist,
                                    df=df, corners=corners)
                    got3 = bandpass(data, low_corner, nyquist + 1.78,
                                    df=df, corners=corners)
                    numwarn = 0
                    for w_ in w:
                        if ('Selected high corner frequency ' in
                                str(w_.message) and
                                'Applying a high-pass instead.' in
                                str(w_.message)):
                            numwarn += 1
                    assert numwarn == 3
                for got in (got1, got2, got3):
                    np.testing.assert_allclose(got, expected, rtol=1e-3,
                                               atol=0.9)

    def test_filterarr2d_axis_and_zerophase(self):
        """
        See #2805
        """
        stream = read()
        df = stream[0].stats.sampling_rate
        arr = np.array([tr.data for tr in stream[:3]])
        for filterf, args in [(bandpass, (1, 5)),
                              (bandstop, (1, 5)),
                              (lowpass, (5,)),
                              (highpass, (5,))]:
            for zerophase in (False, True):
                farray = filterf(arr, *args, df=df, zerophase=zerophase)
                fseries = [filterf(arr[i, :], *args,
                                   df=df, zerophase=zerophase)
                           for i in range(3)]
                assert np.all(np.isclose(farray, fseries))
                # test filter along axis=0
                farray2 = filterf(np.transpose(arr), *args, df=df,
                                  zerophase=zerophase, axis=0)
                assert np.all(np.isclose(farray2, np.transpose(farray)))



# -*- coding: utf-8 -*-
import numpy as np

from obspy import UTCDateTime
from obspy.core.compatibility import from_buffer


def bcd(_i):
    return (_i >> 4 & 0xF).astype(np.uint32) * 10 + (_i & 0xF)


def bcd_16bit_int(_i):
    _i = bcd(_i)
    return _i[::, 0] * 100 + _i[::, 1]


def bcd_hex(_i):
    m = _i.shape[1]
    _bcd = _i.ravel().tobytes().hex().upper()
    return from_buffer(_bcd, dtype="|S%d" % (m * 2))


def bcd_8bit_hex(_i):
    return np.array(["{:X}".format(int(x)) for x in _i], dtype="|S2")


def bcd_julian_day_string_to_nanoseconds_of_year(_i):
    timestrings = bcd_hex(_i)
    return _timestrings_to_nanoseconds(timestrings)


_timegm_cache = {}


def _get_nanoseconds_for_start_of_year(year):
    # Reftek 130 data format stores only the last two digits of the year.
    # We currently assume that 00-49 are years 2000-2049 and 50-99 are years
    # 2050-2099. We deliberately raise an exception in the read routine if the
    # current year will become 2050 (just in case someone really still uses
    # this code then.. ;-)
    # make sure we don't get e.g. np.uint8 input as numpy will not
    # automatically upcast for us with numpy 2.0 anymore
    year = int(year)
    if year < 50:
        year += 2000
    else:
        year += 1900
    try:
        ns = _timegm_cache[year]
    except KeyError:
        ns = UTCDateTime(year, 1, 1)._ns
        _timegm_cache[year] = ns
    return ns


def _timestrings_to_nanoseconds(timestrings):
    """
    Helper routine to convert timestrings of form "DDDHHMMSSsss" to array of
    integer nanosecond POSIX timestamp.

    :param timestring: numpy.ndarray
    :rtype: numpy.ndarray
    """
    # split up the time string into tuple of
    # (day of year, hours, minutes, seconds, milliseconds), still as string
    nanoseconds = [(string[:3], string[3:5], string[5:7],
                    string[7:9], string[9:]) for string in timestrings]
    nanoseconds = np.array(nanoseconds, dtype="S3").astype(np.int64)
    # now scale the columns of the array, so that everything is in nanoseconds
    to_nano = 1000000000
    nanoseconds[:, 0] -= 1
    nanoseconds[:, 0] *= 86400 * to_nano
    nanoseconds[:, 1] *= 3600 * to_nano
    nanoseconds[:, 2] *= 60 * to_nano
    nanoseconds[:, 3] *= to_nano
    nanoseconds[:, 4] *= 1000000
    # sum up days, hours, minutes etc. for every row of the array
    nanoseconds = nanoseconds.sum(axis=1)
    return nanoseconds


def _decode_ascii(chars):
    return chars.decode("ASCII")


def _parse_long_time(time_bytestring, decode=True):
    """
    :returns: POSIX timestamp as integer nanoseconds
    """
    if decode:
        time_string = time_bytestring.decode()
    else:
        time_string = time_bytestring
    if not time_string.strip():
        return None
    time_string, milliseconds = time_string[:-3], int(time_string[-3:])
    t = UTCDateTime.strptime(time_string, '%Y%j%H%M%S')
    nanoseconds = t._ns
    nanoseconds += milliseconds * 1000000
    return nanoseconds


def _16_tuple_ascii(bytestring):
    item_count = 16
    chars = bytestring.decode("ASCII")
    if len(chars) % item_count != 0:
        raise NotImplementedError("Should not happen, contact developers.")
    item_size = int(len(chars) / item_count)
    result = []
    for i in range(item_count):
        chars_ = chars[i * item_size:(i + 1) * item_size]
        result.append(chars_.strip() or None)
    return tuple(result)


def _16_tuple_int(bytestring):
    ascii_tuple = _16_tuple_ascii(bytestring)
    result = []
    for chars in ascii_tuple:
        if chars is None or not chars.strip():
            result.append(None)
            continue
        result.append(int(chars))
    return tuple(result)


def _16_tuple_float(bytestring):
    ascii_tuple = _16_tuple_ascii(bytestring)
    result = []
    for chars in ascii_tuple:
        if chars is None or not chars.strip():
            result.append(None)
            continue
        result.append(float(chars))
    return tuple(result)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
obspy.clients.filesystem.sds - read support for SeisComP Data Structure
=======================================================================
This module provides read support for data stored locally in a SeisComP Data
Structure (SDS) directory structure.

The directory and file layout of SDS is defined as::

    <SDSdir>/YEAR/NET/STA/CHAN.TYPE/NET.STA.LOC.CHAN.TYPE.YEAR.DAY

These fields are defined by SDS as follows::

    SDSdir :  arbitrary base directory
    YEAR   :  4 digit year
    NET    :  Network code/identifier, up to 8 characters, no spaces
    STA    :  Station code/identifier, up to 8 characters, no spaces
    CHAN   :  Channel code/identifier, up to 8 characters, no spaces
    TYPE   :  1 characters indicating the data type, recommended types are:
               'D' - Waveform data
               'E' - Detection data
               'L' - Log data
               'T' - Timing data
               'C' - Calibration data
               'R' - Response data
               'O' - Opaque data
    LOC    :  Location identifier, up to 8 characters, no spaces
    DAY    :  3 digit day of year, padded with zeros

See https://www.seiscomp.de/doc/apps/slarchive.html#slarchive

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import glob
import os
import re
import warnings
from datetime import timedelta

import numpy as np

from obspy import Stream, read, UTCDateTime
from obspy.core.stream import _headonly_warning_msg
from obspy.core.util.misc import BAND_CODE
from obspy.io.mseed import ObsPyMSEEDFilesizeTooSmallError


SDS_FMTSTR = os.path.join(
    "{year}", "{network}", "{station}", "{channel}.{sds_type}",
    "{network}.{station}.{location}.{channel}.{sds_type}.{year}.{doy:03d}")
FORMAT_STR_PLACEHOLDER_REGEX = r"{(\w+?)?([!:].*?)?}"


class Client(object):
    """
    Request client for SeisComP Data Structure archive on local filesystem.

    For details see the :meth:`~obspy.clients.filesystem.sds.Client.__init__()`
    method.
    """
    FMTSTR = SDS_FMTSTR

    def __init__(self, sds_root, sds_type="D", format="MSEED",
                 fileborder_seconds=30, fileborder_samples=5000):
        """
        Initialize a SDS local filesystem client.

        >>> from obspy.clients.filesystem.sds import Client
        >>> client = Client("/my/SDS/archive/root")  # doctest: +SKIP

        :type sds_root: str
        :param sds_root: Root directory of SDS archive.
        :type sds_type: str
        :param sds_type: SDS data type identifier, one single character. Types
            recommended by the SDS definition are: 'D' - Waveform data,
            'E' - Detection data, 'L' - Log data, 'T' - Timing data,
            'C' - Calibration data, 'R' - Response data, 'O' - Opaque data. Can
            also be wildcarded by setting to ``?`` or ``*``.
        :type format: str
        :param format: File format the data is stored in, see
            :func:`~obspy.core.stream.read()` for a list of file formats
            supported by ObsPy. Usually, SDS archives are stored in "MSEED"
            format. Can be set to ``None`` for file format autodetection
            (slowing down the reading).
        :type fileborder_seconds: float
        :param fileborder_seconds: Defines in which cases the client checks the
            previous/next daily file for the requested data (data in SDS
            archives usually spill over on the day break, at least for a few
            seconds). For example setting ``fileborder_seconds=30`` means that
            for a request with ``starttime`` at ``00:00:29`` UTC (or an
            ``endtime`` at ``23:59:31``), the previous daily file is also
            checked, if it contains matching data. The maximum of both
            ``fileborder_seconds`` and ``fileborder_samples`` is used when
            determining if previous/next day should be checked for data.
        :type fileborder_samples: int
        :param fileborder_samples: Similar to ``fileborder_seconds``. The given
            number of samples is converted to seconds by mapping the band
            code of the requested channel to sampling frequency. The maximum of
            both ``fileborder_seconds`` and ``fileborder_samples`` is used when
            determining if previous/next day should be checked for data.
        """
        if not os.path.isdir(sds_root):
            msg = ("SDS root is not a local directory: " + sds_root)
            raise IOError(msg)
        self.sds_root = sds_root
        self.sds_type = sds_type
        self.format = format and format.upper()
        self.fileborder_seconds = fileborder_seconds
        self.fileborder_samples = fileborder_samples

    def get_waveforms(self, network, station, location, channel, starttime,
                      endtime, merge=-1, sds_type=None, **kwargs):
        """
        Read data from a local SeisComP Data Structure (SDS) directory tree.

        >>> from obspy import UTCDateTime
        >>> t = UTCDateTime("2015-10-12T12")
        >>> st = client.get_waveforms("IU", "ANMO", "*", "HH?", t, t+30)
        ... # doctest: +SKIP

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
            Wildcards '*' and '?' are supported.
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
            Wildcards '*' and '?' are supported.
        :type location: str
        :param location: Location code of requested data (e.g. "").
            Wildcards '*' and '?' are supported.
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
            Wildcards '*' and '?' are supported.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        :type merge: int or None
        :param merge: Specifies, which merge operation should be performed
            on the stream before returning the data. Default (``-1``) means
            only a conservative cleanup merge is performed to merge seamless
            traces (e.g. when reading across day boundaries). See
            :meth:`Stream.merge(...) <obspy.core.stream.Stream.merge>` for
            details. If set to ``None`` (or ``False``) no merge operation at
            all will be performed.
        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :param kwargs: Additional kwargs that get passed on to
            :func:`~obspy.core.stream.read` internally, mostly for internal
            low-level purposes used by other methods.
        :rtype: :class:`~obspy.core.stream.Stream`
        """
        if starttime >= endtime:
            msg = ("'endtime' must be after 'starttime'.")
            raise ValueError(msg)
        sds_type = sds_type or self.sds_type

        seed_pattern = ".".join((network, station, location, channel))

        st = Stream()
        full_paths = self._get_filenames(
            network=network, station=station, location=location,
            channel=channel, starttime=starttime, endtime=endtime,
            sds_type=sds_type)
        for full_path in full_paths:
            try:
                st += read(full_path, format=self.format, starttime=starttime,
                           endtime=endtime, sourcename=seed_pattern, **kwargs)
            except ObsPyMSEEDFilesizeTooSmallError:
                # just ignore small MSEED files, in use cases working with
                # near-realtime data these are usually just being created right
                # at request time, e.g. when fetching current data right after
                # midnight
                continue

        # make sure we only have the desired data, just in case the file
        # contents do not match the expected SEED id
        st = st.select(network=network, station=station, location=location,
                       channel=channel)

        # avoid trim/merge operations when we do a headonly read for
        # `_get_availability_percentage()`
        if kwargs.get("_no_trim_or_merge", False):
            return st

        st.trim(starttime, endtime)
        if merge is None or merge is False:
            pass
        else:
            st.merge(merge)
        return st

    def get_waveforms_bulk(self, bulk):
        """
        Reads bulk data from a local SeisComP Data Structure (SDS) directory
        tree.

        Returns a stream object with with the data requested from the local
        directory.

        :type bulk: list[tuple]
        :param bulk: Information about the requested data.
        """
        st = Stream()
        for bulk_string in bulk:
            st += self.get_waveforms(*bulk_string)
        return st

    def _get_filenames(self, network, station, location, channel, starttime,
                       endtime, sds_type=None):
        """
        Get list of filenames for certain waveform and time span.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of time span.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of time span.
        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :rtype: list[str]
        """
        sds_type = sds_type or self.sds_type
        # SDS has data sometimes in adjacent days, so also try to read the
        # requested data from those files. Usually this is only a few seconds
        # of data after midnight, but for now we play safe here to catch all
        # requested data (and with MiniSEED - the usual SDS file format - we
        # can use starttime/endtime kwargs anyway to read only desired parts).
        year_doy = set()
        # determine how far before starttime/after endtime we should check
        # other dayfiles for the data
        t_buffer = self.fileborder_samples / BAND_CODE.get(channel[:1], 20.0)
        t_buffer = max(t_buffer, self.fileborder_seconds)
        t = starttime - t_buffer
        t_max = endtime + t_buffer
        # make a list of year/doy combinations that covers the whole requested
        # time window (plus day before and day after)
        while t < t_max:
            year_doy.add((t.year, t.julday))
            t += timedelta(days=1)
        year_doy.add((t_max.year, t_max.julday))

        full_paths = set()
        for year, doy in year_doy:
            filename = self.FMTSTR.format(
                network=network, station=station, location=location,
                channel=channel, year=year, doy=doy, sds_type=sds_type)
            full_path = os.path.join(self.sds_root, filename)
            full_paths = full_paths.union(glob.glob(full_path))

        return full_paths

    def _get_filename(self, network, station, location, channel, time,
                      sds_type=None):
        """
        Get filename for certain waveform.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type time: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param time: Time of interest.
        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :rtype: str
        """
        sds_type = sds_type or self.sds_type
        filename = self.FMTSTR.format(
            network=network, station=station, location=location,
            channel=channel, year=time.year, doy=time.julday,
            sds_type=sds_type)
        return os.path.join(self.sds_root, filename)

    def get_availability_percentage(self, network, station, location, channel,
                                    starttime, endtime, sds_type=None):
        """
        Get percentage of available data.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start of requested time window.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End of requested time window.
        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :rtype: tuple(float, int)
        :returns: 2-tuple of percentage of available data (``0.0`` to ``1.0``)
            and number of gaps/overlaps.
        """
        if starttime >= endtime:
            msg = ("'endtime' must be after 'starttime'.")
            raise ValueError(msg)
        sds_type = sds_type or self.sds_type
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", _headonly_warning_msg, UserWarning,
                "obspy.core.stream")
            st = self.get_waveforms(network, station, location, channel,
                                    starttime, endtime, sds_type=sds_type,
                                    headonly=True, _no_trim_or_merge=True)
        # even if the warning was silently caught and not shown it gets
        # registered in the __warningregistry__ and will not be shown
        # subsequently in a place were it's not caught
        # see https://bugs.python.org/issue4180
        # see e.g. http://blog.ionelmc.ro/2013/06/26/testing-python-warnings/
        try:
            from obspy.core.stream import __warningregistry__ as \
                stream_warningregistry
        except ImportError:
            # import error means no warning has been issued from
            # obspy.core.stream before, so nothing to do.
            pass
        else:
            for key in list(stream_warningregistry.keys()):
                if key[0] == _headonly_warning_msg:
                    stream_warningregistry.pop(key)
        st.sort(keys=['starttime', 'endtime'])
        st.traces = [tr for tr in st
                     if not (tr.stats.endtime < starttime or
                             tr.stats.starttime > endtime)]

        if not st:
            return (0, 1)

        total_duration = endtime - starttime
        # sum up gaps in the middle
        gaps = [gap[6] for gap in st.get_gaps()]
        gap_sum = np.sum(gaps)
        gap_count = len(gaps)
        # check if we have a gap at start or end
        earliest = min([tr.stats.starttime for tr in st])
        latest = max([tr.stats.endtime for tr in st])
        if earliest > starttime:
            gap_sum += earliest - starttime
            gap_count += 1
        if latest < endtime:
            gap_sum += endtime - latest
            gap_count += 1

        return (1 - (gap_sum / total_duration), gap_count)

    def _get_current_endtime(self, network, station, location, channel,
                             sds_type=None, stop_time=None,
                             check_has_no_data=True):
        """
        Get time of last sample for given stream.

        ``None`` is returned if no data at all is encountered when going
        backwards until `stop_time` (defaults to Jan 1st 1950).

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :type stop_time: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param stop_time: Time at which the search for data is stopped and
            ``None`` is returned. If not specified, stops at ``1950-01-01T00``.
        :type check_has_no_data: bool
        :param check_has_no_data: Whether to perform a check with
            :meth:`has_data` if any data is available at all for the given SEED
            ID. Turns out that this check can take a long time on slow
            filesystems (e.g. NFS), so it might actually make the latency check
            take longer than necessary, so deactivating it might be worth a try
            if speed issues occur.
        :rtype: :class:`~obspy.core.utcdatetime.UTCDateTime` or ``None``
        """
        sds_type = sds_type or self.sds_type

        seed_pattern = ".".join((network, station, location, channel))

        if check_has_no_data and not self.has_data(
                network=network, station=station, location=location,
                channel=channel, sds_type=sds_type):
            return None

        stop_time = stop_time or UTCDateTime(1950, 1, 1)
        st = None
        time = UTCDateTime()

        while not st:
            if time < stop_time:
                return None
            filename = self._get_filename(
                network=network, station=station, location=location,
                channel=channel, time=time, sds_type=sds_type)
            if os.path.isfile(filename):
                try:
                    st = read(filename, format=self.format, headonly=True,
                              sourcename=seed_pattern)
                except ObsPyMSEEDFilesizeTooSmallError:
                    # just ignore small MSEED files, in use cases working with
                    # near-realtime data these are usually just being created
                    # right at request time, e.g. when fetching current data
                    # right after midnight
                    st = None
                else:
                    st = st.select(network=network, station=station,
                                   location=location, channel=channel)
                if st:
                    break
            time -= 24 * 3600

        return max([tr.stats.endtime for tr in st])

    def get_latency(self, network, station, location, channel,
                    sds_type=None, stop_time=None, check_has_no_data=True):
        """
        Get latency for given stream, i.e. difference of current time and
        latest available data for stream in SDS archive. ``None`` is returned
        if no data at all is encountered when going backwards until
        `stop_time` (defaults to Jan 1st 1950).

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :type stop_time: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param stop_time: Time at which the search for data is stopped and
            ``None`` is returned. If not specified, stops at ``1950-01-01T00``.
        :type check_has_no_data: bool
        :param check_has_no_data: Whether to perform a check with
            :meth:`has_data` if any data is available at all for the given SEED
            ID. Turns out that this check can take a long time on slow
            filesystems (e.g. NFS), so it might actually make the latency check
            take longer than necessary, so deactivating it might be worth a try
            if speed issues occur.
        :rtype: float or ``None``
        :returns: Latency in seconds or ``None`` if no data was encountered
            from current time backwards until ``stop_time``.
        """
        endtime = self._get_current_endtime(
            network=network, station=station, location=location,
            channel=channel, sds_type=sds_type, stop_time=stop_time,
            check_has_no_data=check_has_no_data)

        if endtime is None:
            return endtime

        return UTCDateTime() - endtime

    def has_data(self, network, station, location, channel, sds_type=None):
        """
        Check if specified stream has any data.

        Actually just checks whether a file is encountered in a folder that is
        expected to contain data.

        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :rtype: bool
        """
        sds_type = sds_type or self.sds_type

        pattern = re.sub(
            FORMAT_STR_PLACEHOLDER_REGEX,
            _wildcarded_except(["network", "station", "location", "channel",
                                "sds_type"]),
            self.FMTSTR)
        pattern = pattern.format(
            network=network, station=station, location=location,
            channel=channel, sds_type=sds_type)
        pattern = os.path.join(self.sds_root, pattern)
        if glob.glob(pattern):
            return True
        else:
            return False

    def get_all_nslc(self, sds_type=None, datetime=None):
        """
        Return information on what streams are included in archive.

        Note that this can be very slow on network file systems because every
        single file has to be touched (because available location codes can not
        be discovered from folder structure alone).

        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :type datetime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param datetime: Only return all streams that have data at specified
            time (checks if file exists that should have the data, i.e. streams
            might be returned that have data on the same day but not at exactly
            this point in time).
        :rtype: list
        :returns: List of (network, station, location, channel) 4-tuples of all
            available streams in archive.
        """
        sds_type = sds_type or self.sds_type
        result = set()
        # wildcarded pattern to match all files of interest
        if datetime is None:
            pattern = re.sub(
                FORMAT_STR_PLACEHOLDER_REGEX,
                _wildcarded_except(["sds_type"]),
                self.FMTSTR).format(sds_type=sds_type)
            pattern = os.path.join(self.sds_root, pattern)
        else:
            pattern = self._get_filename("*", "*", "*", "*", datetime)
        all_files = glob.glob(pattern)
        # set up inverse regex to extract kwargs/values from full paths
        pattern_ = os.path.join(self.sds_root, self.FMTSTR)
        group_map = {i: groups[0] for i, groups in
                     enumerate(re.findall(FORMAT_STR_PLACEHOLDER_REGEX,
                                          pattern_))}
        for file_ in all_files:
            dict_ = _parse_path_to_dict(file_, pattern_, group_map)
            try:
                network = dict_["network"]
                station = dict_["station"]
                location = dict_["location"]
                channel = dict_["channel"]
            except KeyError as e:
                msg = (
                    "Failed to extract key from pattern '{}' in path "
                    "'{}': {}").format(pattern, file_, e)
                warnings.warn(msg)
                continue
            result.add((network, station, location, channel))
        return sorted(result)

    def get_all_stations(self, sds_type=None):
        """
        Return information on what stations are included in archive.

        This method assumes that network/station combinations can be discovered
        from the folder structure alone (as opposed to the filenames).

        :type sds_type: str
        :param sds_type: Override SDS data type identifier that was specified
            during client initialization.
        :rtype: list
        :returns: List of (network, station) 2-tuples of all available stations
            in archive.
        """
        sds_type = sds_type or self.sds_type
        result = set()
        # wildcarded pattern to match all files of interest
        fmtstr = os.path.dirname(self.FMTSTR)
        pattern = re.sub(
            FORMAT_STR_PLACEHOLDER_REGEX,
            _wildcarded_except(["sds_type"]),
            fmtstr).format(sds_type=sds_type)
        pattern = os.path.join(self.sds_root, pattern)
        all_files = glob.glob(pattern)
        # set up inverse regex to extract kwargs/values from full paths
        pattern_ = os.path.join(self.sds_root, fmtstr)
        group_map = {i: groups[0] for i, groups in
                     enumerate(re.findall(FORMAT_STR_PLACEHOLDER_REGEX,
                                          pattern_))}
        for file_ in all_files:
            dict_ = _parse_path_to_dict(file_, pattern_, group_map)
            try:
                network = dict_["network"]
                station = dict_["station"]
            except (TypeError, KeyError) as e:
                if isinstance(e, TypeError) and dict_ is not None:
                    raise
                msg = (
                    "Failed to extract key from pattern '{}' in path "
                    "'{}': {}").format(pattern_, file_, e)
                warnings.warn(msg)
                continue
            result.add((network, station))
        return sorted(result)


def _wildcarded_except(exclude=[]):
    """
    Function factory for :mod:`re` ``repl`` functions used in :func:`re.sub`,
    replacing all format string place holders with ``*`` wildcards, except
    named fields as specified in ``exclude``.
    """
    def _wildcarded(match):
        if match.group(1) in exclude:
            return match.group(0)
        return "*"
    return _wildcarded


def _parse_path_to_dict(path, pattern, group_map):
    # escape special regex characters "." and "\"
    # in principle we should escape all special characters in Python regex:
    #    . ^ $ * + ? { } [ ] \ | ( )
    # and also only replace them if outside of format string placeholders..
    regex = re.sub(r'([\.\\])', r'\\\1', pattern)
    # replace each format string placeholder with a regex group, matching
    # alphanumerics. append end-of-line otherwise the last non-greedy match
    # doesn't catch anything if it's at the end of the regex
    regex = re.sub(FORMAT_STR_PLACEHOLDER_REGEX, r'(\\w*?)', regex) + "$"
    match = re.match(regex, path)
    if match is None:
        return None
    result = {}
    for i, value in enumerate(match.groups()):
        key = group_map[i]
        if key in result:
            if result[key] != value:
                msg = (
                    "Failed to parse path '{}': Mismatching information "
                    "for key '{}' from pattern '{}'.").format(
                        path, key, pattern)
                warnings.warn(msg)
                return None
        result[key] = value
    return result


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test suite for the channel handling.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2013
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import pytest
from matplotlib import rcParams

from obspy import read_inventory
from obspy.core.inventory import Channel, Equipment
from obspy.core.util.base import CatchAndAssertWarnings


@pytest.mark.usefixtures('ignore_numpy_errors')
class TestChannel:
    """
    Tests the for :class:`~obspy.core.inventory.channel.Channel` class.
    """
    def test_response_plot(self, image_path):
        """
        Tests the response plot.
        """
        cha = read_inventory()[0][0][0]
        with CatchAndAssertWarnings():
            rcParams['savefig.dpi'] = 72
            cha.plot(0.005, outfile=image_path)

    def test_channel_str(self):
        """
        Tests the __str__ method of the channel object.
        """
        c = Channel(code="BHE", location_code="10", latitude=1, longitude=2,
                    elevation=3, depth=4, azimuth=5, dip=6)
        assert str(c) == (
            "Channel 'BHE', Location '10' \n"
            "\tTime range: -- - --\n"
            "\tLatitude: 1.0000, Longitude: 2.0000, Elevation: 3.0 m, "
            "Local Depth: 4.0 m\n"
            "\tAzimuth: 5.00 degrees from north, clockwise\n"
            "\tDip: 6.00 degrees down from horizontal\n")

        # Adding channel types.
        c.types = ["A", "B"]
        assert str(c) == (
            "Channel 'BHE', Location '10' \n"
            "\tTime range: -- - --\n"
            "\tLatitude: 1.0000, Longitude: 2.0000, Elevation: 3.0 m, "
            "Local Depth: 4.0 m\n"
            "\tAzimuth: 5.00 degrees from north, clockwise\n"
            "\tDip: 6.00 degrees down from horizontal\n"
            "\tChannel types: A, B\n")

        # Adding channel types.
        c.sample_rate = 10.0
        assert str(c) == (
            "Channel 'BHE', Location '10' \n"
            "\tTime range: -- - --\n"
            "\tLatitude: 1.0000, Longitude: 2.0000, Elevation: 3.0 m, "
            "Local Depth: 4.0 m\n"
            "\tAzimuth: 5.00 degrees from north, clockwise\n"
            "\tDip: 6.00 degrees down from horizontal\n"
            "\tChannel types: A, B\n"
            "\tSampling Rate: 10.00 Hz\n")

        # "Adding" response
        c.response = True
        assert str(c) == (
            "Channel 'BHE', Location '10' \n"
            "\tTime range: -- - --\n"
            "\tLatitude: 1.0000, Longitude: 2.0000, Elevation: 3.0 m, "
            "Local Depth: 4.0 m\n"
            "\tAzimuth: 5.00 degrees from north, clockwise\n"
            "\tDip: 6.00 degrees down from horizontal\n"
            "\tChannel types: A, B\n"
            "\tSampling Rate: 10.00 Hz\n"
            "\tResponse information available"
        )

        # Adding an empty sensor.
        c.sensor = Equipment(type=None)
        assert str(c) == (
            "Channel 'BHE', Location '10' \n"
            "\tTime range: -- - --\n"
            "\tLatitude: 1.0000, Longitude: 2.0000, Elevation: 3.0 m, "
            "Local Depth: 4.0 m\n"
            "\tAzimuth: 5.00 degrees from north, clockwise\n"
            "\tDip: 6.00 degrees down from horizontal\n"
            "\tChannel types: A, B\n"
            "\tSampling Rate: 10.00 Hz\n"
            "\tSensor (Description): None (None)\n"
            "\tResponse information available"
        )

        # Adding a sensor with only a type.
        c.sensor = Equipment(type="random")
        assert str(c) == (
            "Channel 'BHE', Location '10' \n"
            "\tTime range: -- - --\n"
            "\tLatitude: 1.0000, Longitude: 2.0000, Elevation: 3.0 m, "
            "Local Depth: 4.0 m\n"
            "\tAzimuth: 5.00 degrees from north, clockwise\n"
            "\tDip: 6.00 degrees down from horizontal\n"
            "\tChannel types: A, B\n"
            "\tSampling Rate: 10.00 Hz\n"
            "\tSensor (Description): random (None)\n"
            "\tResponse information available"
        )

        # Adding a sensor with only a description
        c.sensor = Equipment(description="some description")
        assert str(c) == (
            "Channel 'BHE', Location '10' \n"
            "\tTime range: -- - --\n"
            "\tLatitude: 1.0000, Longitude: 2.0000, Elevation: 3.0 m, "
            "Local Depth: 4.0 m\n"
            "\tAzimuth: 5.00 degrees from north, clockwise\n"
            "\tDip: 6.00 degrees down from horizontal\n"
            "\tChannel types: A, B\n"
            "\tSampling Rate: 10.00 Hz\n"
            "\tSensor (Description): None (some description)\n"
            "\tResponse information available"
        )

        # Adding a sensor with type and description
        c.sensor = Equipment(type="random", description="some description")
        assert str(c) == (
            "Channel 'BHE', Location '10' \n"
            "\tTime range: -- - --\n"
            "\tLatitude: 1.0000, Longitude: 2.0000, Elevation: 3.0 m, "
            "Local Depth: 4.0 m\n"
            "\tAzimuth: 5.00 degrees from north, clockwise\n"
            "\tDip: 6.00 degrees down from horizontal\n"
            "\tChannel types: A, B\n"
            "\tSampling Rate: 10.00 Hz\n"
            "\tSensor (Description): random (some description)\n"
            "\tResponse information available"
        )



# -*- coding: utf-8 -*-
"""
obspy.clients.fdsn - FDSN web service client for ObsPy
======================================================
The obspy.clients.fdsn package contains a client to access web servers that
implement the `FDSN web service definitions`_.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)


.. contents:: Contents
    :local:
    :depth: 2

Basic FDSN Client Usage
-----------------------

The first step is always to initialize a client object.

>>> from obspy.clients.fdsn import Client
>>> client = Client("EARTHSCOPE")

A client object can be initialized either with the base URL of any FDSN web
service or with a shortcut name which will be mapped to a FDSN URL. All the
example make use of the FDSN web service at EarthScope (former IRIS). For a
list of other available web service providers, see the
:meth:`~obspy.clients.fdsn.client.Client.__init__()` method. The currently
available providers are:

>>> from obspy.clients.fdsn.header import URL_MAPPINGS
>>> for key in sorted(URL_MAPPINGS.keys()):
...     print("{0:<11} {1}".format(key,  URL_MAPPINGS[key]))  # doctest: +SKIP
AUSPASS     https://auspass.edu.au
BGR         https://eida.bgr.de
BGS         https://eida.bgs.ac.uk
EARTHSCOPE  https://service.earthscope.org
EIDA        http://eida-federator.ethz.ch
EMSC        https://www.seismicportal.eu
EPOSFR      https://seisdata.epos-france.fr
ETH         https://eida.ethz.ch
GEOFON      https://geofon.gfz.de
GEONET      https://service.geonet.org.nz
GFZ         https://geofon.gfz.de
ICGC        https://ws.icgc.cat
IESDMC      http://batsws.earth.sinica.edu.tw
IGN         http://fdsnws.sismologia.ign.es
INGV        https://webservices.ingv.it
IPGP        https://ws.ipgp.fr
IRIS        https://service.earthscope.org
IRISDMC     https://service.earthscope.org
IRISPH5     https://service.earthscope.org
ISC         https://www.isc.ac.uk
KAGSR       http://sdis.emsd.ru
KNMI        https://rdsa.knmi.nl
KOERI       https://eida.koeri.boun.edu.tr
LMU         https://erde.geophysik.uni-muenchen.de
NCEDC       https://service.ncedc.org
NIEP        https://eida-sc3.infp.ro
NOA         https://eida.gein.noa.gr
NRCAN       https://earthquakescanada.nrcan.gc.ca
ODC         https://www.orfeus-eu.org
ORFEUS      https://www.orfeus-eu.org
RASPISHAKE  https://data.raspberryshake.org
RESIF       https://ws.resif.fr
SCEDC       https://service.scedc.caltech.edu
TEXNET      http://rtserve.beg.utexas.edu
UIB-NORSAR  https://eida.geo.uib.no
USGS        https://earthquake.usgs.gov
USP         https://sismo.iag.usp.br

(1) :meth:`~obspy.clients.fdsn.client.Client.get_waveforms()`: The following
    example illustrates how to request and plot 60 minutes of the ``"LHZ"``
    channel of station Albuquerque, New Mexico (``"ANMO"``) of the Global
    Seismograph Network (``"IU"``) for an seismic event around 2010-02-27 06:45
    (UTC). Results are returned as a :class:`~obspy.core.stream.Stream` object.
    See the :meth:`~obspy.clients.fdsn.client.Client.get_waveforms_bulk()`
    method for information on how to send multiple requests simultaneously to
    avoid unnecessary network overhead.

    >>> from obspy import UTCDateTime
    >>> t = UTCDateTime("2010-02-27T06:45:00.000")
    >>> st = client.get_waveforms("IU", "ANMO", "00", "LHZ", t, t + 60 * 60)
    >>> st.plot()  # doctest: +SKIP

    .. plot::

        from obspy import UTCDateTime
        from obspy.clients.fdsn import Client
        client = Client()
        t = UTCDateTime("2010-02-27T06:45:00.000")
        st = client.get_waveforms("IU", "ANMO", "00", "LHZ", t, t + 60 * 60)
        st.plot()

(2) :meth:`~obspy.clients.fdsn.client.Client.get_events()`: Retrieves event
    data from the server. Results are returned as a
    :class:`~obspy.core.event.Catalog` object.

    >>> starttime = UTCDateTime("2002-01-01")
    >>> endtime = UTCDateTime("2002-01-02")
    >>> cat = client.get_events(starttime=starttime, endtime=endtime,
    ...                         minmagnitude=6, catalog="ISC")
    >>> print(cat)  # doctest: +NORMALIZE_WHITESPACE
    2 Event(s) in Catalog:
    2002-01-01T11:29:22.720000Z |  +6.282, +125.749 | 6.3 MW
    2002-01-01T07:28:57.480000Z | +36.991,  +72.336 | 6.3 Mb
    >>> cat.plot()  # doctest: +SKIP

    .. plot::

        from obspy import UTCDateTime
        from obspy.clients.fdsn import Client
        client = Client()
        starttime = UTCDateTime("2002-01-01")
        endtime = UTCDateTime("2002-01-02")
        cat = client.get_events(starttime=starttime, endtime=endtime,
                                minmagnitude=6, catalog="ISC")
        cat.plot()

(3) :meth:`~obspy.clients.fdsn.client.Client.get_stations()`: Retrieves station
    data from the server. Results are returned as an
    :class:`~obspy.core.inventory.inventory.Inventory` object.

    >>> inventory = client.get_stations(network="IU", station="A*",
    ...                                 starttime=starttime,
    ...                                 endtime=endtime)
    >>> print(inventory)  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Inventory created at ...
        Created by: IRIS WEB SERVICE: fdsnws-station | version: ...
                    ...
        Sending institution: IRIS-DMC (IRIS-DMC)
        Contains:
                Networks (1):
                        IU
                Stations (3):
                        IU.ADK (Adak, Aleutian Islands, Alaska)
                        IU.AFI (Afiamalu, Samoa)
                        IU.ANMO (Albuquerque, New Mexico, USA)
                Channels (0):
    >>> inventory.plot()  # doctest: +SKIP

    .. plot::

        from obspy import UTCDateTime
        from obspy.clients.fdsn import Client
        client = Client()
        starttime = UTCDateTime("2002-01-01")
        endtime = UTCDateTime("2002-01-02")
        inventory = client.get_stations(network="IU", station="A*",
                                        starttime=starttime,
                                        endtime=endtime)
        inventory.plot()


Basic Routing Clients Usage
---------------------------

Routers are web services that can be queried for which data centers offer
certain pieces of data. That information can then be used to get the actual
data from various data center across the globe. All current routing services
only support the ``dataselect`` and the ``station`` FDSNWS services.

ObsPy has support for two routing services:

(i) The `EarthScope (former IRIS) Federator
    <https://service.earthscope.org/irisws/fedcatalog/1/>`_.
(ii) The `EIDAWS Routing Service
     <http://www.orfeus-eu.org/data/eida/webservices/routing/>`_.

To use them, call the
:func:`~obspy.clients.fdsn.routing.routing_client.RoutingClient` function:


>>> from obspy.clients.fdsn import RoutingClient

Get an instance of a routing client using the EarthScope Federator:

>>> client = RoutingClient("earthscope-federator")
>>> print(type(client))  # doctest: +ELLIPSIS
<class '...fdsn.routing.federator_routing_client.FederatorRoutingClient'>

Or get an instance of a routing client using the EIDAWS routing web service:

>>> client = RoutingClient("eida-routing")
>>> print(type(client))  # doctest: +ELLIPSIS
<class '...fdsn.routing.eidaws_routing_client.EIDAWSRoutingClient'>

They can be used like the normal FDSNWS clients, meaning the
``get_(waveforms|stations)(_bulk)()`` functions should work as expected.

To be able to do geographic waveform queries with the EIDA service,
ObsPy will internally perform a station query before downloading the
waveforms. This results in a similar usage between the EIDA and EarthScope
(former IRIS) routing services from a user's perspective.

The following snippet will call the EarthScope federator to figure out who
has waveform data for that particular query and subsequently call the
individual data centers to actually get the data. This happens fully
automatically - please note that the clients also supports non-standard
waveform query parameters like geographical constraints.

>>> from obspy import UTCDateTime
>>> client = RoutingClient("earthscope-federator")
>>> st = client.get_waveforms(
...     channel="LHZ", starttime=UTCDateTime(2017, 1, 1),
...     endtime=UTCDateTime(2017, 1, 1, 0, 5), latitude=10,
...     longitude=10, maxradius=25)  # doctest: +SKIP
>>> print(st)  # doctest: +SKIP
2 Trace(s) in Stream:
II.MBAR.00.LHZ | 2017-01-01T00:00:00Z - ... | 1.0 Hz, 300 samples
II.MBAR.10.LHZ | 2017-01-01T00:00:00Z - ... | 1.0 Hz, 300 samples

The same works for stations:

>>> client = RoutingClient("earthscope-federator")
>>> inv = client.get_stations(
...     channel="LHZ", starttime=UTCDateTime(2017, 1, 1),
...     endtime=UTCDateTime(2017, 1, 1, 0, 5), latitude=10,
...     level="channel", longitude=10, maxradius=25)  # doctest: +SKIP
>>> print(inv)  # doctest: +SKIP
Inventory created at 2017-10-09T14:26:28.466161Z
    Created by: ObsPy 1.0.3.post0+1706.gef068de324.dirty
                https://www.obspy.org
    Sending institution: IRIS-DMC,ObsPy ..., SeisComP3 (GFZ,IPGP, IRIS-DMC)
    Contains:
        Networks (6):
            AF, G, II, IU, TT, YY
        Stations (10):
            AF.EKNA (Ekona, Cameroon)
            AF.KIG (Kigali, Rwanda)
            G.TAM (Tamanrasset, Algeria)
            II.MBAR (Mbarara, Uganda)
            IU.KOWA (Kowa, Mali)
            TT.TATN (Station Tataouine, Tunisia)
            YY.GIDA (Health Center, Gidami, Ethiopia)
            YY.GUBA (Police Station, Guba, Ethiopia)
            YY.MEND (High School, Mendi, Ethiopia)
            YY.SHER (Police Station, Sherkole, Ethiopia)
        Channels (11):
            AF.EKNA..LHZ, AF.KIG..LHZ, G.TAM.00.LHZ, II.MBAR.00.LHZ,
            II.MBAR.10.LHZ, IU.KOWA.00.LHZ, TT.TATN.00.LHZ, YY.GIDA..LHZ,
            YY.GUBA..LHZ, YY.MEND..LHZ, YY.SHER..LHZ

Please see the documentation for each method for further information and
examples.

.. _FDSN web service definitions: https://www.fdsn.org/webservices/
"""
from .client import Client  # NOQA
from .routing.routing_client import RoutingClient  # NOQA
from .header import URL_MAPPINGS  # NOQA


# insert supported URL mapping list dynamically in docstring
# we need an if clause because add_doctests() executes the file once again
if Client.__init__.__doc__ is not None and r"%s" in Client.__init__.__doc__:
    Client.__init__.__doc__ = \
        Client.__init__.__doc__ % \
        str(sorted(URL_MAPPINGS.keys())).strip("[]")

__all__ = ["Client", "RoutingClient"]


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



from __future__ import print_function

from math import log10

import numpy as np

from obspy.clients.arclink import Client
from obspy import Stream, UTCDateTime
from obspy.geodetics import gps2dist_azimuth
from obspy.signal.trigger import coincidence_trigger


client = Client(user="sed-workshop@obspy.org")

t = UTCDateTime("2012-04-03T01:00:00")
t2 = t + 4 * 3600

stations = ["AIGLE", "SENIN", "DIX", "LAUCH", "MMK", "SIMPL"]
st = Stream()

for station in stations:
    try:
        tmp = client.get_waveforms("CH", station, "", "[EH]HZ", t, t2,
                                   metadata=True)
    except Exception:
        print(station, "---")
        continue
    st += tmp

st.taper()
st.filter("bandpass", freqmin=1, freqmax=20)
triglist = coincidence_trigger("recstalta", 10, 2, st, 4, sta=0.5, lta=10)
print(len(triglist), "events triggered.")

for trig in triglist:
    closest_sta = trig['stations'][0]
    tr = st.select(station=closest_sta)[0]
    trig['latitude'] = tr.stats.coordinates.latitude
    trig['longitude'] = tr.stats.coordinates.longitude

paz_wa = {'sensitivity': 2800, 'zeros': [0j], 'gain': 1,
          'poles': [-6.2832 - 4.7124j, -6.2832 + 4.7124j]}

for trig in triglist:
    t = trig['time']
    print("#" * 80)
    print("Trigger time:", t)
    mags = []

    stations = client.get_stations(t, t + 300, "CH")

    for station in stations:
        station = station['code']
        try:
            st = client.get_waveforms("CH", station, "", "[EH]H[ZNE]", t - 300,
                                      t + 300, metadata=True)
            assert len(st) == 3
        except Exception:
            print(station, "---")
            continue

        st.simulate(paz_remove="self", paz_simulate=paz_wa, water_level=10)
        st.trim(t, t + 50)

        tr_n = st.select(component="N")[0]
        ampl_n = max(abs(tr_n.data))
        tr_e = st.select(component="E")[0]
        ampl_e = max(abs(tr_e.data))
        ampl = max(ampl_n, ampl_e)

        sta_lat = st[0].stats.coordinates.latitude
        sta_lon = st[0].stats.coordinates.longitude
        event_lat = trig['latitude']
        event_lon = trig['longitude']

        epi_dist, az, baz = gps2dist_azimuth(event_lat, event_lon, sta_lat,
                                             sta_lon)
        epi_dist = epi_dist / 1000

        if epi_dist < 60:
            a = 0.018
            b = 2.17
        else:
            a = 0.0038
            b = 3.02
        ml = log10(ampl * 1000) + a * epi_dist + b
        print(station, ml)
        mags.append(ml)

    net_mag = np.median(mags)
    print("Network magnitude:", net_mag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The obspy.clients.fdsn.client test suite.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import io
import re
import socket
import sys
import warnings
from difflib import Differ
from unittest import mock
from urllib.error import URLError

import urllib.request as urllib_request

import lxml
import numpy as np
import pytest
import requests

from obspy import (
    UTCDateTime, read, read_inventory, Stream, Trace, Inventory, Catalog)
from obspy.core.util.base import NamedTemporaryFile, CatchAndAssertWarnings
from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning
from obspy.clients.fdsn import Client, RoutingClient
from obspy.clients.fdsn.client import (build_url, parse_simple_xml,
                                       get_bulk_string, _cleanup_earthscope,
                                       raise_on_error)
from obspy.clients.fdsn.header import (DEFAULT_USER_AGENT, URL_MAPPINGS,
                                       FDSNException, FDSNRedirectException,
                                       FDSNNoDataException,
                                       FDSNRequestTooLargeException,
                                       FDSNBadRequestException,
                                       FDSNNoAuthenticationServiceException,
                                       FDSNTimeoutException,
                                       FDSNNoServiceException,
                                       FDSNInternalServerException,
                                       FDSNTooManyRequestsException,
                                       FDSNNotImplementedException,
                                       FDSNBadGatewayException,
                                       FDSNServiceUnavailableException,
                                       FDSNUnauthorizedException,
                                       FDSNForbiddenException,
                                       FDSNDoubleAuthenticationException,
                                       FDSNInvalidRequestException,
                                       DEFAULT_SERVICES)
from obspy.core.inventory import Response
from obspy.geodetics import locations2degrees


USER_AGENT = "ObsPy (test suite) " + " ".join(DEFAULT_USER_AGENT.split())


def _normalize_stats(obj):
    if isinstance(obj, Stream):
        for tr in obj:
            _normalize_stats(tr)
    else:
        if "processing" in obj.stats:
            del obj.stats["processing"]
        if "_fdsnws_dataselect_url" in obj.stats:
            del obj.stats._fdsnws_dataselect_url


def failmsg(got, expected, ignore_lines=[]):
    """
    Create message on difference between objects.

    If both are strings create a line-by-line diff, otherwise create info on
    both using str().
    For diffs, lines that contain any string given in ignore_lines will be
    excluded from the comparison.
    """
    if isinstance(got, str) and isinstance(expected, str):
        got = [line for line in got.splitlines(True)
               if all([x not in line for x in ignore_lines])]
        expected = [line for line in expected.splitlines(True)
                    if all([x not in line for x in ignore_lines])]
        diff = Differ().compare(got, expected)
        diff = "".join([line for line in diff if line[0] in "-+?"])
        if diff:
            return "\nDiff:\n%s" % diff
        else:
            return ""
    else:
        return "\nGot:\n%s\nExpected:\n%s" % (str(got), str(expected))


def normalize_version_number(string):
    """
    Returns imput string with version numbers normalized for testing purposes.
    """
    match = r'v[0-9]+\.[0-9]+\.[0-9]+'
    repl = re.sub(match, "vX.X.X", string).replace(",", "")
    return [line.strip() for line in repl.splitlines()]


@pytest.mark.network
class TestClient():
    """
    Test cases for obspy.clients.fdsn.client.Client.
    """
    @classmethod
    def setup_class(cls):
        cls.client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT)
        cls.client_auth = \
            Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                   user="nobody@iris.edu", password="anonymous")

    # @pytest.mark.skip(reason='data no longer available')
    def test_trim_stream_after_get_waveform(self):
        """
        Tests that stream is properly trimmed to user requested times after
        fetching from datacenter, see #1887
        """
        c = Client(
            service_mappings={'dataselect':
                              'https://ws.ipgp.fr/fdsnws/dataselect/1'})
        starttime = UTCDateTime('2016-11-01T00:00:00')
        endtime = UTCDateTime('2016-11-01T00:00:10')
        stream = c.get_waveforms('G', 'PEL', '*', 'LHZ', starttime, endtime)
        assert starttime == stream[0].stats.starttime
        assert endtime == stream[0].stats.endtime

    def test_service_discovery_iris(self):
        """
        Tests the automatic discovery of services with the EARTHSCOPE endpoint.
        The test parameters are taken from EARTHSCOPE website.

        This will have to be adjusted once EARTHSCOPE changes their
        implementation.
        """
        client = self.client
        assert {*client.services.keys()} == \
            {"dataselect", "event", "station", "available_event_contributors",
             "available_event_catalogs"}

        # The test sets are copied from the EARTHSCOPE webpage.
        assert {*client.services["dataselect"].keys()} == \
            {"starttime", "endtime", "network", "station", "location",
             "channel"}
        assert {*client.services["station"].keys()} == \
            {"starttime", "endtime", "startbefore", "startafter",
             "endbefore", "endafter", "network", "station", "location",
             "channel", "minlatitude", "maxlatitude", "minlongitude",
             "maxlongitude", "latitude", "longitude", "minradius",
             "maxradius", "level", "includerestricted", "format",
             "includeavailability", "updatedafter", "matchtimeseries"}
        assert {*client.services["event"].keys()} == \
            {"starttime", "endtime", "minlatitude", "maxlatitude",
             "minlongitude", "maxlongitude", "latitude", "longitude",
             "maxradius", "minradius", "mindepth", "maxdepth",
             "minmagnitude", "maxmagnitude",
             "magnitudetype", "format",
             "catalog", "contributor", "limit", "offset", "orderby",
             "updatedafter", "includeallorigins", "includeallmagnitudes",
             "includearrivals", "eventid",
             "originid"}  # XXX: This is currently just specified in the WADL

        # Also check an exemplary value in more detail.
        minradius = client.services["event"]["minradius"]
        assert minradius["default_value"] == 0.0
        assert not minradius["required"]
        assert minradius["doc"] == ""
        assert minradius["doc_title"] == (
            "Specify minimum distance from the geographic point defined by "
            "latitude and longitude")
        assert minradius["type"] == float
        assert minradius["options"] == []

    def test_iris_event_catalog_availability(self):
        """
        Tests the parsing of the available event catalogs.
        """
        assert {*self.client.services["available_event_catalogs"]} == \
            {"GCMT", "ISC", "NEIC PDE"}

    def test_iris_event_contributors_availability(self):
        """
        Tests the parsing of the available event contributors.
        """
        response = requests.get(
            'https://service.earthscope.org/fdsnws/event/1/contributors')
        xml = lxml.etree.fromstring(response.content)
        expected = {
            elem.text for elem in xml.xpath('/Contributors/Contributor')}
        expected = set(_cleanup_earthscope(expected))
        # check that we have some values in there
        assert len(expected) > 5
        assert {*self.client.services["available_event_contributors"]} == \
            expected

    def test_iris_example_queries_event(self):
        """
        Tests the (sometimes modified) example queries given on the EARTHSCOPE
        web page.

        Used to be tested against files but that was not maintainable. It
        now tests if the queries return what was asked for.
        """
        client = self.client

        # Event id query.
        cat = client.get_events(eventid=609301)
        assert len(cat) == 1
        assert "609301" in cat[0].resource_id.id

        # Temporal query.
        cat = client.get_events(
            starttime=UTCDateTime("2001-01-07T01:00:00"),
            endtime=UTCDateTime("2001-01-07T01:05:00"), catalog="ISC")
        assert len(cat) > 0
        for event in cat:
            assert event.origins[0].extra.catalog.value == "ISC"
            assert event.origins[0].time > UTCDateTime("2001-01-07T01:00:00")
            assert UTCDateTime("2001-01-07T01:05:00") > event.origins[0].time

        # Misc query.
        cat = client.get_events(
            starttime=UTCDateTime("2001-01-07T14:00:00"),
            endtime=UTCDateTime("2001-01-08T00:00:00"), minlatitude=15,
            maxlatitude=40, minlongitude=-170, maxlongitude=170,
            includeallmagnitudes=True, minmagnitude=4, orderby="magnitude")
        assert len(cat) > 0
        for event in cat:
            assert event.origins[0].time > \
                               UTCDateTime("2001-01-07T14:00:00")
            assert UTCDateTime("2001-01-08T00:00:00") > event.origins[0].time
            assert event.origins[0].latitude > 14.9
            assert 40.1 > event.origins[0].latitude
            assert event.origins[0].latitude > -170.1
            assert 170.1 > event.origins[0].latitude
            # events returned by FDSNWS can contain many magnitudes with a wide
            # range, and currently (at least for EARTHSCOPE) the magnitude
            # threshold sent to the server checks if at least one magnitude
            # matches, it does not only check the preferred magnitude..
            assert any(m.mag >= 3.999 for m in event.magnitudes)

    @pytest.mark.filterwarnings('ignore:.*cannot deal with')
    def test_irisph5_event(self):
        """
        Tests the IRISPH5 URL mapping, which is special due to its custom
        subpath.
        """
        client = Client('IRISPH5')

        # Event id query.
        cat = client.get_events(catalog='8A')
        assert len(cat) == 19
        assert cat[0].event_type == 'controlled explosion'

    def test_iris_example_queries_station(self):
        """
        Tests the (sometimes modified) example queries given on EARTHSCOPE
        webpage.

        This test used to download files but that is almost impossible to
        keep up to date - thus it is now a bit smarter and tests the
        returned inventory in different ways.
        """
        client = self.client

        # Radial query.
        inv = client.get_stations(latitude=-56.1, longitude=-26.7,
                                  maxradius=15)
        assert len(inv.networks) > 0  # at least one network
        for net in inv:
            assert len(net.stations) > 0  # at least one station
            for sta in net:
                dist = locations2degrees(sta.latitude, sta.longitude,
                                         -56.1, -26.7)
                # small tolerance for WGS84.
                assert 15.1 > dist, "%s.%s" % (net.code, sta.code)

        # Misc query.
        inv = client.get_stations(
            startafter=UTCDateTime("2003-01-07"),
            endbefore=UTCDateTime("2011-02-07"), minlatitude=15,
            maxlatitude=55, minlongitude=170, maxlongitude=-170, network="IM")
        assert len(inv.networks) > 0  # at least one network
        for net in inv:
            assert len(net.stations) > 0  # at least one station
            for sta in net:
                msg = "%s.%s" % (net.code, sta.code)
                assert sta.start_date > UTCDateTime("2003-01-07"), msg
                if sta.end_date is not None:
                    assert UTCDateTime("2011-02-07") > sta.end_date, \
                                       msg
                assert sta.latitude > 14.9, msg
                assert 55.1 > sta.latitude, msg
                assert not (-170.1 <= sta.longitude <= 170.1), msg
                assert net.code == "IM", msg

        # Simple query
        inv = client.get_stations(
            starttime=UTCDateTime("2000-01-01"),
            endtime=UTCDateTime("2001-01-01"), net="IU", sta="ANMO")
        assert len(inv.networks) > 0  # at least one network
        for net in inv:
            assert len(net.stations) > 0  # at least one station
            for sta in net:
                assert UTCDateTime("2001-01-01") > sta.start_date
                if sta.end_date is not None:
                    assert sta.end_date > UTCDateTime("2000-01-01")
                assert net.code == "IU"
                assert sta.code == "ANMO"

        # Station wildcard query.
        inv = client.get_stations(
            starttime=UTCDateTime("2000-01-01"),
            endtime=UTCDateTime("2002-01-01"), network="IU", sta="A*",
            location="00")
        assert len(inv.networks) > 0  # at least one network
        for net in inv:
            assert len(net.stations) > 0  # at least one station
            for sta in net:
                assert UTCDateTime("2002-01-01") > sta.start_date
                if sta.end_date is not None:
                    assert sta.end_date > UTCDateTime("2000-01-01")
                assert net.code == "IU"
                assert sta.code.startswith("A")

    def test_iris_example_queries_dataselect(self, testdata):
        """
        Tests the (sometimes modified) example queries given on EARTHSCOPE
        webpage.
        """
        client = self.client

        queries = [
            ("IU", "ANMO", "00", "BHZ",
             UTCDateTime("2010-02-27T06:30:00.000"),
             UTCDateTime("2010-02-27T06:40:00.000")),
            ("IU", "A*", "*", "BHZ",
             UTCDateTime("2010-02-27T06:30:00.000"),
             UTCDateTime("2010-02-27T06:31:00.000")),
            ("IU", "A??", "*0", "BHZ",
             UTCDateTime("2010-02-27T06:30:00.000"),
             UTCDateTime("2010-02-27T06:31:00.000")),
        ]
        result_files = ["dataselect_example.mseed",
                        "dataselect_example_wildcards.mseed",
                        "dataselect_example_mixed_wildcards.mseed",
                        ]
        for query, filename in zip(queries, result_files):
            # test output to stream
            got = client.get_waveforms(*query)
            # Assert that the meta-information about the provider is stored.
            for tr in got:
                assert tr.stats._fdsnws_dataselect_url == \
                    client.base_url + "/fdsnws/dataselect/1/query"
            # Remove fdsnws URL as it is not in the data from the disc.
            for tr in got:
                del tr.stats._fdsnws_dataselect_url
            expected = read(testdata[filename])
            # The client trims by default.
            _normalize_stats(got)
            assert got == expected, \
                "Dataselect failed for query %s" % repr(query)
            # test output to file
            with NamedTemporaryFile() as tf:
                client.get_waveforms(*query, filename=tf.name)
                with open(tf.name, 'rb') as fh:
                    got = fh.read()
                with open(testdata[filename], 'rb') as fh:
                    expected = fh.read()
            assert got == expected, \
                "Dataselect failed for query %s" % repr(query)

    def test_authentication(self, testdata):
        """
        Test dataselect with authentication.
        """
        client = self.client_auth
        # dataselect example queries
        query = ("IU", "ANMO", "00", "BHZ",
                 UTCDateTime("2010-02-27T06:30:00.000"),
                 UTCDateTime("2010-02-27T06:40:00.000"))
        got = client.get_waveforms(*query)
        expected = read(testdata["dataselect_example.mseed"])
        _normalize_stats(got)
        assert got == expected, failmsg(got, expected)

    def test_iris_example_queries_event_discover_services_false(self):
        """
        Tests the (sometimes modified) example queries given on the EARTHSCOPE
        web page, without service discovery.

        Used to be tested against files but that was not maintainable. It
        now tests if the queries return what was asked for.
        """
        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        _discover_services=False)

        # Event id query.
        cat = client.get_events(eventid=609301)
        assert len(cat) == 1
        assert "609301" in cat[0].resource_id.id

        # Temporal query.
        cat = client.get_events(
            starttime=UTCDateTime("2001-01-07T01:00:00"),
            endtime=UTCDateTime("2001-01-07T01:05:00"), catalog="ISC")
        assert len(cat) > 0
        for event in cat:
            assert event.origins[0].extra.catalog.value == "ISC"
            assert event.origins[0].time > UTCDateTime("2001-01-07T01:00:00")
            assert UTCDateTime("2001-01-07T01:05:00") > event.origins[0].time

        # Misc query.
        cat = client.get_events(
            starttime=UTCDateTime("2001-01-07T14:00:00"),
            endtime=UTCDateTime("2001-01-08T00:00:00"), minlatitude=15,
            maxlatitude=40, minlongitude=-170, maxlongitude=170,
            includeallmagnitudes=True, minmagnitude=4, orderby="magnitude")
        assert len(cat) > 0
        for event in cat:
            assert event.origins[0].time > \
                               UTCDateTime("2001-01-07T14:00:00")
            assert UTCDateTime("2001-01-08T00:00:00") > event.origins[0].time
            assert event.origins[0].latitude > 14.9
            assert 40.1 > event.origins[0].latitude
            assert event.origins[0].latitude > -170.1
            assert 170.1 > event.origins[0].latitude
            # events returned by FDSNWS can contain many magnitudes with a wide
            # range, and currently (at least for EARTHSCOPE) the magnitude
            # threshold sent to the server checks if at least one magnitude
            # matches, it does not only check the preferred magnitude..
            assert any(m.mag >= 3.999 for m in event.magnitudes)

    def test_iris_example_queries_station_discover_services_false(self):
        """
        Tests the (sometimes modified) example queries given on EARTHSCOPE
        webpage, without service discovery.

        This test used to download files but that is almost impossible to
        keep up to date - thus it is now a bit smarter and tests the
        returned inventory in different ways.
        """
        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        _discover_services=False)

        # Radial query.
        inv = client.get_stations(latitude=-56.1, longitude=-26.7,
                                  maxradius=15)
        assert len(inv.networks) > 0  # at least one network
        for net in inv:
            assert len(net.stations) > 0  # at least one station
            for sta in net:
                dist = locations2degrees(sta.latitude, sta.longitude,
                                         -56.1, -26.7)
                # small tolerance for WGS84.
                assert 15.1 > dist, "%s.%s" % (net.code, sta.code)

        # Misc query.
        inv = client.get_stations(
            startafter=UTCDateTime("2003-01-07"),
            endbefore=UTCDateTime("2011-02-07"), minlatitude=15,
            maxlatitude=55, minlongitude=170, maxlongitude=-170, network="IM")
        assert len(inv.networks) > 0  # at least one network
        for net in inv:
            assert len(net.stations) > 0  # at least one station
            for sta in net:
                msg = "%s.%s" % (net.code, sta.code)
                assert sta.start_date > UTCDateTime("2003-01-07"), msg
                if sta.end_date is not None:
                    assert UTCDateTime("2011-02-07") > sta.end_date, \
                                       msg
                assert sta.latitude > 14.9, msg
                assert 55.1 > sta.latitude, msg
                assert not (-170.1 <= sta.longitude <= 170.1), msg
                assert net.code == "IM", msg

        # Simple query
        inv = client.get_stations(
            starttime=UTCDateTime("2000-01-01"),
            endtime=UTCDateTime("2001-01-01"), net="IU", sta="ANMO")
        assert len(inv.networks) > 0  # at least one network
        for net in inv:
            assert len(net.stations) > 0  # at least one station
            for sta in net:
                assert UTCDateTime("2001-01-01") > sta.start_date
                if sta.end_date is not None:
                    assert sta.end_date > UTCDateTime("2000-01-01")
                assert net.code == "IU"
                assert sta.code == "ANMO"

        # Station wildcard query.
        inv = client.get_stations(
            starttime=UTCDateTime("2000-01-01"),
            endtime=UTCDateTime("2002-01-01"), network="IU", sta="A*",
            location="00")
        assert len(inv.networks) > 0  # at least one network
        for net in inv:
            assert len(net.stations) > 0  # at least one station
            for sta in net:
                assert UTCDateTime("2002-01-01") > sta.start_date
                if sta.end_date is not None:
                    assert sta.end_date > UTCDateTime("2000-01-01")
                assert net.code == "IU"
                assert sta.code.startswith("A")

    def test_iris_example_queries_dataselect_discover_services_false(
            self, testdata):
        """
        Tests the (sometimes modified) example queries given on EARTHSCOPE
        webpage, without discovering services first.
        """
        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        _discover_services=False)

        queries = [
            ("IU", "ANMO", "00", "BHZ",
             UTCDateTime("2010-02-27T06:30:00.000"),
             UTCDateTime("2010-02-27T06:40:00.000")),
            ("IU", "A*", "*", "BHZ",
             UTCDateTime("2010-02-27T06:30:00.000"),
             UTCDateTime("2010-02-27T06:31:00.000")),
            ("IU", "A??", "*0", "BHZ",
             UTCDateTime("2010-02-27T06:30:00.000"),
             UTCDateTime("2010-02-27T06:31:00.000")),
        ]
        result_files = ["dataselect_example.mseed",
                        "dataselect_example_wildcards.mseed",
                        "dataselect_example_mixed_wildcards.mseed",
                        ]
        for query, filename in zip(queries, result_files):
            # test output to stream
            got = client.get_waveforms(*query)
            # Assert that the meta-information about the provider is stored.
            for tr in got:
                assert tr.stats._fdsnws_dataselect_url == \
                    client.base_url + "/fdsnws/dataselect/1/query"
            # Remove fdsnws URL as it is not in the data from the disc.
            for tr in got:
                del tr.stats._fdsnws_dataselect_url
            expected = read(testdata[filename])
            _normalize_stats(got)
            assert got == expected, \
                "Dataselect failed for query %s" % repr(query)
            # test output to file
            with NamedTemporaryFile() as tf:
                client.get_waveforms(*query, filename=tf.name)
                with open(tf.name, 'rb') as fh:
                    got = fh.read()
                with open(testdata[filename], 'rb') as fh:
                    expected = fh.read()
            assert got == expected, \
                "Dataselect failed for query %s" % repr(query)

    def test_help_function_with_earthscope(self, testdata):
        """
        Tests the help function with the EARTHSCOPE example.

        This will have to be adopted any time EARTHSCOPE changes their
        implementation.
        """
        try:
            client = self.client

            # Capture output
            tmp = io.StringIO()
            sys.stdout = tmp

            client.help("event")
            got = sys.stdout.getvalue()
            sys.stdout = sys.__stdout__
            tmp.close()
            filename = "event_helpstring.txt"
            with open(testdata[filename]) as fh:
                expected = fh.read()
            # allow for changes in version number..
            got = normalize_version_number(got)
            expected = normalize_version_number(expected)
            # catalogs/contributors are checked in separate tests
            assert got[-2].startswith('Available catalogs:')
            assert got[-1].startswith('Available contributors:')
            got = got[:-2]
            expected = expected[:-2]
            for line_got, line_expected in zip(got, expected):
                assert line_got == line_expected

            # Reset. Creating a new one is faster then clearing the old one.
            tmp = io.StringIO()
            sys.stdout = tmp

            client.help("station")
            got = sys.stdout.getvalue()
            sys.stdout = sys.__stdout__
            tmp.close()

            filename = "station_helpstring.txt"
            with open(testdata[filename]) as fh:
                expected = fh.read()
            got = normalize_version_number(got)
            expected = normalize_version_number(expected)
            assert got == expected, failmsg(got, expected)

            # Reset.
            tmp = io.StringIO()
            sys.stdout = tmp

            client.help("dataselect")
            got = sys.stdout.getvalue()
            sys.stdout = sys.__stdout__
            tmp.close()

            filename = "dataselect_helpstring.txt"
            with open(testdata[filename]) as fh:
                expected = fh.read()
            got = normalize_version_number(got)
            expected = normalize_version_number(expected)
            assert got == expected, failmsg(got, expected)

        finally:
            sys.stdout = sys.__stdout__

    def test_str_method(self):
        got = str(self.client)
        expected = (
            "FDSN Webservice Client "
            "(base url: https://service.earthscope.org)\n"
            "Available Services: 'dataselect' (v1.0.0), 'event' (v1.0.6), "
            "'station' (v1.0.7), 'available_event_catalogs', "
            "'available_event_contributors'\n\n"
            "Use e.g. client.help('dataselect') for the\n"
            "parameter description of the individual services\n"
            "or client.help() for parameter description of\n"
            "all webservices.")
        got = normalize_version_number(got)
        expected = normalize_version_number(expected)
        assert got == expected, failmsg(got, expected)

    @pytest.mark.parametrize('auth', (True, False))
    def test_dataselect_bulk(self, testdata, auth):
        """
        Test bulk dataselect requests, POSTing data to server. Also tests
        authenticated bulk request.
        """
        if auth:
            client = self.client_auth
        else:
            client = self.client

        expected = read(testdata["bulk.mseed"])
        # test cases for providing lists of lists
        # Deliberately requesting data that overlap the end-time of a channel.
        # TA.A25A..BHZ ends at 2011-07-22T14:50:25.5
        bulk = (("TA", "A25A", "", "BHZ",
                 UTCDateTime("2011-07-22T14:50:23"),
                 UTCDateTime("2011-07-22T14:50:29")),
                ("TA", "A25A", "", "BHE",
                 UTCDateTime("2010-03-25T00:00:00"),
                 UTCDateTime("2010-03-25T00:00:06")),
                ("IU", "ANMO", "*", "HHZ",
                 UTCDateTime("2010-03-25T00:00:00"),
                 UTCDateTime("2010-03-25T00:00:08")))
        # removed the params (min length etc, the test will fail)
        # test output to stream
        got = client.get_waveforms_bulk(bulk)
        # Remove fdsnws URL as it is not in the data from the disc.
        for tr in got:
            del tr.stats._fdsnws_dataselect_url
        assert got == expected, failmsg(got, expected)
        # test output to file
        with NamedTemporaryFile() as tf:
            client.get_waveforms_bulk(bulk, filename=tf.name)
            got = read(tf.name)
        assert got == expected, failmsg(got, expected)
        # test cases for providing a request string
        # note that Earthscope now ignores all those three parameters we used
        # to send along for testing, all of "quality", "longestonly" and
        # "minimumlength" gets ignored now
        bulk = ("quality=B\n"
                "longestonly=false\n"
                "minimumlength=5\n"
                "TA A25A -- BHZ 2011-07-22T14:50:23 2011-07-22T14:50:29\n"
                "TA A25A -- BHE 2010-03-25T00:00:00 2010-03-25T00:00:06\n"
                "IU ANMO * HHZ 2010-03-25T00:00:00 2010-03-25T00:00:08\n")
        # test output to stream
        got = client.get_waveforms_bulk(bulk)
        # Assert that the meta-information about the provider is stored.
        for tr in got:
            if client.user:
                assert tr.stats._fdsnws_dataselect_url == \
                    client.base_url + "/fdsnws/dataselect/1/queryauth"
            else:
                assert tr.stats._fdsnws_dataselect_url == \
                    client.base_url + "/fdsnws/dataselect/1/query"
        # Remove fdsnws URL as it is not in the data from the disc.
        for tr in got:
            del tr.stats._fdsnws_dataselect_url
        assert got == expected, failmsg(got, expected)
        # test output to file
        with NamedTemporaryFile() as tf:
            client.get_waveforms_bulk(bulk, filename=tf.name)
            got = read(tf.name)
        assert got == expected, failmsg(got, expected)
        # test cases for providing a file name
        with NamedTemporaryFile() as tf:
            with open(tf.name, "wt") as fh:
                fh.write(bulk)
            got = client.get_waveforms_bulk(bulk)
        # Remove fdsnws URL as it is not in the data from the disc.
        for tr in got:
            del tr.stats._fdsnws_dataselect_url
        assert got == expected, failmsg(got, expected)
        # test cases for providing a file-like object
        got = client.get_waveforms_bulk(io.StringIO(bulk))
        # Remove fdsnws URL as it is not in the data from the disc.
        for tr in got:
            del tr.stats._fdsnws_dataselect_url
        assert got == expected, failmsg(got, expected)

    def test_station_bulk(self):
        """
        Test bulk station requests, POSTing data to server. Also tests
        authenticated bulk request.

        Does currently only test reading from a list of list. The other
        input types are tested with the waveform bulk downloader and thus
        should work just fine.
        """
        clients = [self.client, self.client_auth]
        # test cases for providing lists of lists
        starttime = UTCDateTime(1990, 1, 1)
        endtime = UTCDateTime(1990, 1, 1) + 10
        bulk = [
            ["IU", "ANMO", "", "BHE", starttime, endtime],
            ["IU", "CCM", "", "BHZ", starttime, endtime],
            ["IU", "COR", "", "UHZ", starttime, endtime],
            ["IU", "HRV", "", "LHN", starttime, endtime],
        ]
        for client in clients:
            # Test with station level.
            inv = client.get_stations_bulk(bulk, level="station")
            # Test with output to file.
            with NamedTemporaryFile() as tf:
                client.get_stations_bulk(
                    bulk, filename=tf.name, level="station")
                inv2 = read_inventory(tf.name, format="stationxml")

            assert inv.networks == inv2.networks
            assert len(inv.networks) == 1
            assert inv[0].code == "IU"
            assert len(inv.networks[0].stations) == 4
            assert sorted([_i.code for _i in inv.networks[0].stations]) == \
                sorted(["ANMO", "CCM", "COR", "HRV"])

            # Test with channel level.
            inv = client.get_stations_bulk(bulk, level="channel")
            # Test with output to file.
            with NamedTemporaryFile() as tf:
                client.get_stations_bulk(
                    bulk, filename=tf.name, level="channel")
                inv2 = read_inventory(tf.name, format="stationxml")

            assert inv.networks == inv2.networks
            assert len(inv.networks) == 1
            assert inv[0].code == "IU"
            assert len(inv.networks[0].stations) == 4
            assert sorted([_i.code for _i in inv.networks[0].stations]) == \
                sorted(["ANMO", "CCM", "COR", "HRV"])
            channels = []
            for station in inv[0]:
                for channel in station:
                    channels.append("IU.%s.%s.%s" % (
                        station.code, channel.location_code,
                        channel.code))
            assert sorted(channels) == \
                sorted(["IU.ANMO..BHE", "IU.CCM..BHZ", "IU.COR..UHZ",
                        "IU.HRV..LHN"])
        return

    def test_get_waveform_attach_response(self):
        """
        minimal test for automatic attaching of metadata
        """
        client = self.client

        bulk = ("IU ANMO 00 BHZ 2000-03-25T00:00:00 2000-03-25T00:00:04\n")
        st = client.get_waveforms_bulk(bulk, attach_response=True)
        for tr in st:
            assert isinstance(tr.stats.get("response"), Response)

        st = client.get_waveforms("IU", "ANMO", "00", "BHZ",
                                  UTCDateTime("2000-02-27T06:00:00.000"),
                                  UTCDateTime("2000-02-27T06:00:05.000"),
                                  attach_response=True)
        for tr in st:
            assert isinstance(tr.stats.get("response"), Response)

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_default_requested_urls(self, download_url_mock):
        """
        Five request should be sent upon initializing a client. Test these.
        """
        download_url_mock.return_value = (404, None)
        base_url = "http://example.com"

        # An exception will be raised if not actual WADLs are returned.
        try:
            Client(base_url=base_url)
        except FDSNException:
            pass

        expected_urls = sorted([
            "%s/fdsnws/event/1/contributors" % base_url,
            "%s/fdsnws/event/1/catalogs" % base_url,
            "%s/fdsnws/event/1/application.wadl" % base_url,
            "%s/fdsnws/station/1/application.wadl" % base_url,
            "%s/fdsnws/dataselect/1/application.wadl" % base_url,
        ])
        got_urls = sorted([_i[0][0] for _i in
                           download_url_mock.call_args_list])

        assert expected_urls == got_urls

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_setting_service_major_version(self, download_url_mock):
        """
        Test the setting of custom major versions.
        """
        download_url_mock.return_value = (404, None)
        base_url = "http://example.com"

        # Passing an empty dictionary results in the default urls.
        major_versions = {}
        # An exception will be raised if not actual WADLs are returned.
        try:
            Client(base_url=base_url, major_versions=major_versions)
        except FDSNException:
            pass
        expected_urls = sorted([
            "%s/fdsnws/event/1/contributors" % base_url,
            "%s/fdsnws/event/1/catalogs" % base_url,
            "%s/fdsnws/event/1/application.wadl" % base_url,
            "%s/fdsnws/station/1/application.wadl" % base_url,
            "%s/fdsnws/dataselect/1/application.wadl" % base_url,
        ])
        got_urls = sorted([_i[0][0] for _i in
                           download_url_mock.call_args_list])
        assert expected_urls == got_urls

        # Replace all
        download_url_mock.reset_mock()
        download_url_mock.return_value = (404, None)
        major_versions = {"event": 7, "station": 8, "dataselect": 9}
        # An exception will be raised if not actual WADLs are returned.
        try:
            Client(base_url=base_url, major_versions=major_versions)
        except FDSNException:
            pass
        expected_urls = sorted([
            "%s/fdsnws/event/7/contributors" % base_url,
            "%s/fdsnws/event/7/catalogs" % base_url,
            "%s/fdsnws/event/7/application.wadl" % base_url,
            "%s/fdsnws/station/8/application.wadl" % base_url,
            "%s/fdsnws/dataselect/9/application.wadl" % base_url,
        ])
        got_urls = sorted([_i[0][0] for _i in
                           download_url_mock.call_args_list])
        assert expected_urls == got_urls

        # Replace only some
        download_url_mock.reset_mock()
        download_url_mock.return_value = (404, None)
        major_versions = {"event": 7, "station": 8}
        # An exception will be raised if not actual WADLs are returned.
        try:
            Client(base_url=base_url, major_versions=major_versions)
        except FDSNException:
            pass
        expected_urls = sorted([
            "%s/fdsnws/event/7/contributors" % base_url,
            "%s/fdsnws/event/7/catalogs" % base_url,
            "%s/fdsnws/event/7/application.wadl" % base_url,
            "%s/fdsnws/station/8/application.wadl" % base_url,
            "%s/fdsnws/dataselect/1/application.wadl" % base_url,
        ])
        got_urls = sorted([_i[0][0] for _i in
                           download_url_mock.call_args_list])
        assert expected_urls == got_urls

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_setting_service_provider_mappings(self, download_url_mock):
        """
        Tests the setting of per service endpoints
        """
        base_url = "http://example.com"

        # Replace all.
        download_url_mock.return_value = (404, None)
        # Some custom urls
        base_url_event = "http://other_url.com/beta/event_service/11"
        base_url_station = "http://some_url.com/beta2/stat_serv/7"
        base_url_ds = "http://new.com/beta3/waveforms/8"
        # An exception will be raised if not actual WADLs are returned.
        try:
            Client(base_url=base_url, service_mappings={
                "event": base_url_event,
                "station": base_url_station,
                "dataselect": base_url_ds,
            })
        except FDSNException:
            pass
        expected_urls = sorted([
            "%s/contributors" % base_url_event,
            "%s/catalogs" % base_url_event,
            "%s/application.wadl" % base_url_event,
            "%s/application.wadl" % base_url_station,
            "%s/application.wadl" % base_url_ds,
        ])
        got_urls = sorted([_i[0][0] for _i in
                           download_url_mock.call_args_list])
        assert expected_urls == got_urls

        # Replace only two. The others keep the default mapping.
        download_url_mock.reset_mock()
        download_url_mock.return_value = (404, None)
        # Some custom urls
        base_url_station = "http://some_url.com/beta2/stat_serv/7"
        base_url_ds = "http://new.com/beta3/waveforms/8"
        # An exception will be raised if not actual WADLs are returned.
        try:
            Client(base_url=base_url, service_mappings={
                "station": base_url_station,
                "dataselect": base_url_ds,
            })
        except FDSNException:
            pass
        expected_urls = sorted([
            "%s/fdsnws/event/1/contributors" % base_url,
            "%s/fdsnws/event/1/catalogs" % base_url,
            "%s/fdsnws/event/1/application.wadl" % base_url,
            "%s/application.wadl" % base_url_station,
            "%s/application.wadl" % base_url_ds,
        ])
        got_urls = sorted([_i[0][0] for _i in
                           download_url_mock.call_args_list])
        assert expected_urls == got_urls

    def test_manually_deactivate_single_service(self):
        """
        Test manually deactivating a single service.
        """
        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        service_mappings={"event": None})
        assert sorted(client.services.keys()) == ['dataselect', 'station']

    def test_redirection(self):
        """
        Tests the redirection of GET and POST requests. We redirect
        everything if not authentication is used.

        EARTHSCOPE runs three services to test it:
            https://ds.iris.edu/files/redirect/307/station/1
            https://ds.iris.edu/files/redirect/307/dataselect/1
            https://ds.iris.edu/files/redirect/307/event/1
        """
        c = Client("EARTHSCOPE", service_mappings={
            "station":
                "https://ds.iris.edu/files/redirect/307/station/1",
            "dataselect":
                "https://ds.iris.edu/files/redirect/307/dataselect/1",
            "event":
                "https://ds.iris.edu/files/redirect/307/event/1"},
            user_agent=USER_AGENT)

        st = c.get_waveforms(
            network="IU", station="ANMO", location="00", channel="BHZ",
            starttime=UTCDateTime("2010-02-27T06:30:00.000"),
            endtime=UTCDateTime("2010-02-27T06:30:01.000"))
        # Just make sure something is being downloaded.
        assert bool(len(st))

        inv = c.get_stations(
            starttime=UTCDateTime("2000-01-01"),
            endtime=UTCDateTime("2001-01-01"),
            network="IU", station="ANMO", level="network")
        # Just make sure something is being downloaded.
        assert bool(len(inv.networks))

        cat = c.get_events(starttime=UTCDateTime("2001-01-07T01:00:00"),
                           endtime=UTCDateTime("2001-01-07T01:05:00"),
                           catalog="ISC")
        # Just make sure something is being downloaded.
        assert bool(len(cat))

        # Also test the bulk requests which are done using POST requests.
        bulk = (("TA", "A25A", "", "BHZ",
                 UTCDateTime("2010-03-25T00:00:00"),
                 UTCDateTime("2010-03-25T00:00:01")),
                ("TA", "A25A", "", "BHE",
                 UTCDateTime("2010-03-25T00:00:00"),
                 UTCDateTime("2010-03-25T00:00:01")))
        st = c.get_waveforms_bulk(bulk, quality="B", longestonly=False)
        # Just make sure something is being downloaded.
        assert bool(len(st))

        starttime = UTCDateTime(1990, 1, 1)
        endtime = UTCDateTime(1990, 1, 1) + 10
        bulk = [
            ["IU", "ANMO", "", "BHE", starttime, endtime],
            ["IU", "CCM", "", "BHZ", starttime, endtime],
        ]
        inv = c.get_stations_bulk(bulk, level="network")
        # Just make sure something is being downloaded.
        assert bool(len(inv.networks))

    def test_redirection_auth(self):
        """
        Tests the redirection of GET and POST requests using authentication.

        By default these should not redirect and an exception is raised.
        """
        # Clear the cache.
        Client._Client__service_discovery_cache.clear()

        # The error will already be raised during the initialization in most
        # cases.
        service_mappings = {
            "station": (
                "https://ds.iris.edu/files/redirect/307/station/1"
            ),
            "dataselect": (
                "https://ds.iris.edu/files/redirect/307/dataselect/1"
            ),
            "event": "https://ds.iris.edu/files/redirect/307/event/1"}
        with warnings.catch_warnings():
            # ignore warnings about unclosed sockets
            # These occur when rasing the FDSNRedirectException, but
            # I was not able to fix in the code
            warnings.filterwarnings('ignore', 'unclosed')
            with pytest.raises(FDSNRedirectException):
                Client("EARTHSCOPE", service_mappings=service_mappings,
                       user="nobody@earthscope.org", password="anonymous",
                       user_agent=USER_AGENT)
            # The force_redirect flag overwrites that behaviour.
            c_auth = Client("EARTHSCOPE", service_mappings=service_mappings,
                            user="nobody@earthscope.org",
                            password="anonymous",
                            user_agent=USER_AGENT, force_redirect=True)
        st = c_auth.get_waveforms(
            network="IU", station="ANMO", location="00", channel="BHZ",
            starttime=UTCDateTime("2010-02-27T06:30:00.000"),
            endtime=UTCDateTime("2010-02-27T06:30:01.000"))
        # Just make sure something is being downloaded.
        assert bool(len(st))

        inv = c_auth.get_stations(
            starttime=UTCDateTime("2000-01-01"),
            endtime=UTCDateTime("2001-01-01"),
            network="IU", station="ANMO", level="network")
        # Just make sure something is being downloaded.
        assert bool(len(inv.networks))

        cat = c_auth.get_events(starttime=UTCDateTime("2001-01-07T01:00:00"),
                                endtime=UTCDateTime("2001-01-07T01:05:00"),
                                catalog="ISC")
        # Just make sure something is being downloaded.
        assert bool(len(cat))

        # Also test the bulk requests which are done using POST requests.
        bulk = (("TA", "A25A", "", "BHZ",
                 UTCDateTime("2010-03-25T00:00:00"),
                 UTCDateTime("2010-03-25T00:00:01")),
                ("TA", "A25A", "", "BHE",
                 UTCDateTime("2010-03-25T00:00:00"),
                 UTCDateTime("2010-03-25T00:00:01")))
        st = c_auth.get_waveforms_bulk(bulk, quality="B", longestonly=False)
        # Just make sure something is being downloaded.
        assert bool(len(st))

        starttime = UTCDateTime(1990, 1, 1)
        endtime = UTCDateTime(1990, 1, 1) + 10
        bulk = [
            ["IU", "ANMO", "", "BHE", starttime, endtime],
            ["IU", "CCM", "", "BHZ", starttime, endtime],
        ]
        inv = c_auth.get_stations_bulk(bulk, level="network")
        # Just make sure something is being downloaded.
        assert bool(len(inv.networks))

    def test_get_waveforms_empty_seed_codes(self):
        """
        Make sure that network, station, and channel codes specified as empty
        strings are not omitted in `get_waveforms(...)` when building the url
        (which results in default values '*' (wildcards) at the server,
        see #1578).
        """
        t = UTCDateTime(2000, 1, 1)
        url_base = "https://service.earthscope.org/fdsnws/dataselect/1/query?"
        kwargs = dict(network='IU', station='ANMO', location='00',
                      channel='HHZ', starttime=t, endtime=t)

        for key in ('network', 'station', 'channel'):
            kwargs_ = kwargs.copy()
            # set empty SEED code for given key
            kwargs_.update(((key, ''),))

            # use a mock object and check what URL would have been downloaded
            with mock.patch.object(
                    self.client, '_download') as m:
                try:
                    self.client.get_waveforms(**kwargs_)
                except Exception:
                    # Mocking returns something different.
                    continue
                # URL downloading comes before the error and can be checked now
                url = m.call_args[0][0]
            url_parts = url.replace(url_base, '').split("&")
            assert '{}='.format(key) in url_parts

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_authentication_exceptions(self, download_url_mock):
        """
        Verify that a request with missing authentication raises an
        identifiable exception
        """
        with mock.patch("obspy.clients.fdsn.client.Client._has_eida_auth",
                        new_callable=mock.PropertyMock,
                        return_value=False):
            with pytest.raises(FDSNNoAuthenticationServiceException):
                Client("GFZ", eida_token="TEST", _discover_services=False)

        with pytest.raises(FDSNDoubleAuthenticationException):
            Client("GFZ", eida_token="TEST", user="TEST",
                   _discover_services=False)

        download_url_mock.reset_mock()
        download_url_mock.return_value = (401, None)
        with pytest.raises(FDSNUnauthorizedException):
            self.client.get_stations(network='00', station='123')

        download_url_mock.return_value = (403, None)
        with pytest.raises(FDSNForbiddenException):
            self.client.get_stations(network='00', station='123')

    def test_no_service_exception(self):
        """
        Verify that opening a client to a provider without FDSN service
        raises an identifiable exception
        """
        with pytest.raises(FDSNNoServiceException):
            Client("http://nofdsnservice.org")

    @pytest.mark.skip(reason='EIDA token is expired')
    def test_eida_token_resolution(self, testdata):
        """
        Tests that EIDA tokens are resolved correctly and new credentials get
        installed with the opener of the Client.

        """
        token = testdata['eida_token.txt']
        with open(token, 'rb') as fh:
            token_data = fh.read().decode()

        def _assert_eida_user_and_password(user, password):
            # user/pass is not static for the static test token
            for value in user, password:
                # seems safe to assume both user and password are at least 10
                # chars long
                # example user/password:
                # wWGgJnH4GvdVY7gDMH21xEpb wDnzlpljqdaCXlP2
                re.match('^[a-zA-Z0-9]{10,}$', value)

        def _get_http_digest_auth_handler(client):
            handlers = [h for h in client._url_opener.handlers
                        if isinstance(h, urllib_request.HTTPDigestAuthHandler)]
            assert len(handlers) <= 1
            return handlers and handlers[0] or None

        def _assert_credentials(client, user, password):
            handler = _get_http_digest_auth_handler(client)
            assert isinstance(handler, urllib_request.HTTPDigestAuthHandler)
            for user_, password_ in handler.passwd.passwd[None].values():
                assert user == user_
                assert password == password_

        client = Client('GFZ')
        # this is a plain client, so it should not have http digest auth
        assert _get_http_digest_auth_handler(client) is None
        # now, if we set new user/password, we should get a http digest auth
        # handler
        user, password = ("spam", "eggs")
        client._set_opener(user=user, password=password)
        _assert_credentials(client, user, password)
        # now, if we resolve the EIDA token, the http digest auth handler
        # should change
        user, password = client._resolve_eida_token(token=token)
        _assert_eida_user_and_password(user, password)
        client._set_opener(user=user, password=password)
        _assert_credentials(client, user, password)
        # do it again, now providing the token data directly as a string (first
        # change the authentication again to dummy user/password
        client._set_opener(user="foo", password="bar")
        _assert_credentials(client, "foo", "bar")
        user, password = client._resolve_eida_token(token=token_data)
        _assert_eida_user_and_password(user, password)
        client.set_eida_token(token_data)
        _assert_credentials(client, user, password)

        # Raise if token and user/pw are given.
        msg = "EIDA authentication token provided, but user and password " \
              "are also given."
        with pytest.raises(FDSNException, match=msg):
            Client('GFZ', eida_token=token, user="foo", password="bar")

        # now lets test the RoutingClient with credentials..
        credentials_ = {'geofon.gfz.de': {'eida_token': token}}
        credentials_mapping_ = {'GFZ': {'eida_token': token}}
        global_eida_credentials_ = {'EIDA_TOKEN': token}
        for credentials, should_have_credentials in zip(
                (None, credentials_, credentials_mapping_,
                 global_eida_credentials_), (False, True, True, True)):
            def side_effect(self_, *args, **kwargs):
                """
                This mocks out Client.get_waveforms_bulk which gets called by
                the routing client, checks authentication handlers and returns
                a dummy stream.
                """
                # check that we're at the expected FDSN WS server
                assert 'https://geofon.gfz.de' == self_.base_url
                # check if credentials were used
                # eida auth availability should be positive in all cases
                assert self_._has_eida_auth
                # depending on whether we specified credentials, the
                # underlying FDSN client should have EIDA authentication
                # flag and should also have a HTTP digest handler with
                # appropriate user/password
                handler = _get_http_digest_auth_handler(self_)
                if should_have_credentials:
                    for user, password in handler.passwd.passwd[None].values():
                        _assert_eida_user_and_password(user, password)
                else:
                    assert handler is None
                # just always return some dummy stream, we're not
                # interested in checking the data downloading which
                # succeeds regardless if auth is used or not as it's public
                # data
                return Stream([Trace(data=np.ones(2))])

            with mock.patch(
                    'obspy.clients.fdsn.client.Client.get_waveforms_bulk',
                    autospec=True) as p:

                p.side_effect = side_effect

                routing_client = RoutingClient('eida-routing',
                                               credentials=credentials)
                # do a waveform request on the routing client which internally
                # connects to the GFZ FDSNWS. this should be done using the
                # above supplied credentials, i.e. should use the given EIDA
                # token to resolve user/password for the normal FDSN queryauth
                # mechanism
                routing_client.get_waveforms(
                    network="GE", station="KMBO", location="00", channel="BHZ",
                    starttime=UTCDateTime("2010-02-27T06:30:00.000"),
                    endtime=UTCDateTime("2010-02-27T06:40:00.000"))

        # test invalid token/token file
        with pytest.raises(
                ValueError,
                match='EIDA token does not seem to be a valid PGP message'):
            client = Client('GFZ', eida_token="spam")
        msg = ("Read EIDA token from file '[^']*event_helpstring.txt' but it "
               "does not seem to contain a valid PGP message.")
        with pytest.raises(ValueError, match=msg):
            client = Client('GFZ', eida_token=testdata['event_helpstring.txt'])


class TestClientNoNetwork():
    """
    Test cases for obspy.clients.fdsn.client.Client that do not need network
    access.
    """
    @classmethod
    def setup_class(cls):
        cls.client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                            _discover_services=False)
        cls.client_auth = \
            Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                   user="nobody@earthscope.org", password="anonymous",
                   _discover_services=False)

    def test_empty_bulk_string(self):
        """
        Makes sure an exception is raised if an empty bulk string would be
        produced (e.g. empty list as input for `get_bulk_string()`)
        """
        msg = ("Empty 'bulk' parameter potentially leading to a FDSN request "
               "of all available data")
        for bad_input in [[], '', None]:
            with pytest.raises(FDSNInvalidRequestException, match=msg):
                get_bulk_string(bulk=bad_input, arguments={})

    def test_validate_base_url(self):
        """
        Tests the _validate_base_url() method.
        """

        test_urls_valid = list(URL_MAPPINGS.values())
        test_urls_valid += [
            "http://something.ethz.ch",
            "http://example.org",
            "https://webservices.rm.ingv.it",
            "http://localhost:8080/test/",
            "http://93.63.40.85/",
            "http://[::1]:80/test/",
            "http://[2001:db8:85a3:8d3:1319:8a2e:370:7348]",
            "http://[2001:db8::ff00:42:8329]",
            "http://[::ffff:192.168.89.9]",
            "http://jane",
            "http://localhost",
            "http://hyphenated-internal-hostname",
            "http://internal-machine.private",
            "https://long-public-tld.international",
            "http://punycode-tld.xn--fiqs8s"]

        test_urls_fails = [
            "http://",
            "http://127.0.1",
            "http://127.=.0.1",
            "http://127.0.0.0.1",
            "http://tld.too.long." + ("x" * 64)]
        test_urls_fails += [
            "http://[]",
            "http://[1]",
            "http://[1:2]",
            "http://[1::2::3]",
            "http://[1::2:3::4]",
            "http://[1:2:2:4:5:6:7]"]

        for url in test_urls_valid:
            assert self.client._validate_base_url(url), \
                '%s should be valid' % url

        for url in test_urls_fails:
            assert not self.client._validate_base_url(url), \
                '%s should be invalid' % url

    def test_url_building(self):
        """
        Tests the build_url() functions.
        """
        # Application WADL
        base = "https://service.earthscope.org"

        assert build_url(base, "dataselect", 1, "application.wadl") == (
            f"{base}/fdsnws/dataselect/1/application.wadl"
        )
        assert build_url(base, "event", 1, "application.wadl") == (
            f"{base}/fdsnws/event/1/application.wadl"
        )
        assert build_url(base, "station", 1, "application.wadl") == (
            f"{base}/fdsnws/station/1/application.wadl"
        )
        # Test one parameter.
        url = build_url(base, "dataselect", 1, "query", {"network": "BW"})
        assert url == f"{base}/fdsnws/dataselect/1/query?network=BW"

        url = build_url(base, "dataselect", 1, "queryauth", {"network": "BW"})
        assert url == f"{base}/fdsnws/dataselect/1/queryauth?network=BW"

        # Test two parameters. Note random order, two possible results.
        params = {"net": "A", "sta": "BC"}
        url = build_url(base, "dataselect", 1, "query", params)
        assert url in (
            f"{base}/fdsnws/dataselect/1/query?net=A&sta=BC",
            f"{base}/fdsnws/dataselect/1/query?sta=BC&net=A",
        )

        # A wrong service raises a ValueError
        with pytest.raises(ValueError):
            build_url("https://service.earthscope.org", "obspy", 1, "query")

    def test_location_parameters(self):
        """
        Tests how the variety of location values are handled.

        Why location? Mostly because it is one tricky parameter. It is not
        uncommon to assume that a non-existent location is "--", but in reality
        "--" is "<space><space>". This substitution exists because mostly
        because various applications have trouble digesting spaces (spaces in
        the URL, for example).
        The confusion begins when location is treated as empty instead, which
        would imply "I want all locations" instead of "I only want locations of
        <space><space>"
        """
        base = "https://service.earthscope.org"
        base_station = "https://service.earthscope.org/fdsnws/station"

        # requests with no specified location should be treated as a wildcard
        params = {"network": "IU", "station": "ANMO",
                  "starttime": "2013-01-01"}
        url = build_url(base, "station", 1, "query", params)
        assert "--" not in url

        # location of "  " is the same as "--"
        url = build_url(base, "station", 1, "query", {"location": "  "})
        assert url == f"{base_station}/1/query?location=--"

        # wildcard locations are valid. Will be encoded.
        url = build_url(base, "station", 1, "query", {"location": "*"})
        assert url == f"{base_station}/1/query?location=%2A"

        url = build_url(base, "station", 1, "query", {"location": "A?"})
        assert url == f"{base_station}/1/query?location=A%3F"

        # lists are valid, including <space><space> lists.
        # Again encoded result
        url = build_url(base, "station", 1, "query", {"location": "  ,1?,?0"})
        assert url == f"{base_station}/1/query?location=--%2C1%3F%2C%3F0"

        url = build_url(base, "station", 1, "query", {"location": "1?,--,?0"})
        assert url == f"{base_station}/1/query?location=1%3F%2C--%2C%3F0"

        # Test all three special cases with empty parameters into lists.
        url = build_url(base, "station", 1, "query", {"location": "  ,AA,BB"})
        assert url == f"{base_station}/1/query?location=--%2CAA%2CBB"

        url = build_url(base, "station", 1, "query", {"location": "AA,  ,BB"})
        assert url == f"{base_station}/1/query?location=AA%2C--%2CBB"

        url = build_url(base, "station", 1, "query", {"location": "AA,BB,  "})
        assert url == f"{base_station}/1/query?location=AA%2CBB%2C--"

        # The location parameter is also passed through the
        # _create_url_from_parameters() method and thus has to survive it!
        # This guards against a regression where all empty location codes
        # where removed by this function!
        for service in ["station", "dataselect"]:
            for loc in ["", " ", "  ", "--", b"", b" ", b"  ", b"--",
                        u"", u" ", u"  ", u"--"]:
                assert "location=--" in \
                    self.client._create_url_from_parameters(
                        service, [],
                        {"location": loc, "starttime": 0, "endtime": 1})

        # Also check the full call with a mock test.
        for loc in ["", " ", "  ", "--", b"", b" ", b"  ", b"--",
                    u"", u" ", u"  ", u"--"]:
            with mock.patch("obspy.clients.fdsn.Client._download") as p:
                self.client.get_stations(0, 0, location=loc,
                                         filename=mock.Mock())
            assert p.call_count == 1
            assert "location=--" in p.call_args[0][0]
            with mock.patch("obspy.clients.fdsn.Client._download") as p:
                self.client.get_waveforms(1, 2, loc, 4, 0, 0,
                                          filename=mock.Mock())
            assert p.call_count == 1
            assert "location=--" in p.call_args[0][0]

    def test_url_building_with_auth(self):
        """
        Tests the Client._build_url() method with authentication.

        Necessary on top of test_url_building test case because clients with
        authentication have to build different URLs for dataselect.
        """
        base = "https://service.earthscope.org"

        # no authentication
        got = self.client._build_url("dataselect", "query", {'net': "BW"})
        expected = f"{base}/fdsnws/dataselect/1/query?net=BW"
        assert got == expected
        # with authentication
        got = self.client_auth._build_url("dataselect", "query", {'net': "BW"})
        expected = (f"{base}/fdsnws/dataselect/1/"
                    "queryauth?net=BW")
        assert got == expected

    def test_set_credentials(self):
        """
        Test for issue #2146

        When setting credentials not during `__init__` but using
        `set_credentials`, waveform queries should still properly go to
        "queryauth" endpoint.
        """
        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        _discover_services=False)
        user = "nobody@earthscope.org"
        password = "anonymous"
        client.set_credentials(user=user, password=password)
        got = client._build_url("dataselect", "query", {'net': "BW"})
        expected = ("https://service.earthscope.org/fdsnws/dataselect/1/"
                    "queryauth?net=BW")
        assert got == expected
        # more basic test: check that set_credentials has set Client.user
        # (which is tested when checking which endpoint to use, query or
        # queryauth)
        assert client.user == user

    def test_discover_services_defaults(self):
        """
        A Client initialized with _discover_services=False shouldn't perform
        any services/WADL queries on the endpoint, and should show only the
        default service parameters.
        """
        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        _discover_services=False)
        assert client.services == DEFAULT_SERVICES

    def test_simple_xml_parser(self):
        """
        Tests the simple XML parsing helper function.
        """
        catalogs = parse_simple_xml("""
            <?xml version="1.0"?>
            <Catalogs>
              <total>6</total>
              <Catalog>ANF</Catalog>
              <Catalog>GCMT</Catalog>
              <Catalog>TEST</Catalog>
              <Catalog>ISC</Catalog>
              <Catalog>UofW</Catalog>
              <Catalog>NEIC PDE</Catalog>
            </Catalogs>""")
        assert catalogs == {"catalogs": set(
            ("ANF", "GCMT", "TEST", "ISC", "UofW", "NEIC PDE"))}

    def test_conflicting_params(self):
        """
        """
        with pytest.raises(FDSNInvalidRequestException):
            self.client.get_stations(network="IU", net="IU")

    @mock.patch(
        "obspy.clients.fdsn.client.Client._get_webservice_versionstring")
    def test_str_method(self, version_mock):
        # doesn't matter what version a server would actually return, since
        # we normalize the version numbers during testing the string anyway
        version_mock.return_value = '1.1.9'
        got = str(self.client)
        expected = (
            "FDSN Webservice Client "
            "(base url: https://service.earthscope.org)\n"
            "Available Services: 'dataselect' (v1.0.0), 'event' (v1.0.6), "
            "'station' (v1.0.7)\n\n"
            "Use e.g. client.help('dataselect') for the\n"
            "parameter description of the individual services\n"
            "or client.help() for parameter description of\n"
            "all webservices.")
        got = normalize_version_number(got)
        expected = normalize_version_number(expected)
        assert got == expected, failmsg(got, expected)

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_download_urls_for_custom_mapping(
            self, download_url_mock, testdata):
        """
        Tests the downloading of data with custom mappings.
        """
        base_url = "http://example.com"

        # More extensive mock setup simulation service discovery.
        def custom_side_effects(*args, **kwargs):
            if "version" in args[0]:
                return 200, "1.0.200"
            elif "event" in args[0]:
                with open(testdata["2014-01-07_iris_event.wadl"], "rb") as fh:
                    return 200, fh.read()
            elif "station" in args[0]:
                with open(testdata["2014-01-07_iris_station.wadl"],
                          "rb") as fh:
                    return 200, fh.read()
            elif "dataselect" in args[0]:
                with open(testdata["2014-01-07_iris_dataselect.wadl"],
                          "rb") as fh:
                    return 200, fh.read()
            return 404, None

        download_url_mock.side_effect = custom_side_effects

        # Some custom urls
        base_url_event = "http://example.com/beta/event_service/11"
        base_url_station = "http://example.org/beta2/station/7"
        base_url_ds = "http://example.edu/beta3/dataselect/8"

        # An exception will be raised if not actual WADLs are returned.
        # Catch warnings to avoid them being raised for the tests.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            c = Client(base_url=base_url, service_mappings={
                "event": base_url_event,
                "station": base_url_station,
                "dataselect": base_url_ds,
            })
        for warning in w:
            assert "Could not parse" in str(warning) or \
                            "cannot deal with" in str(warning)

        # Test the dataselect downloading.
        download_url_mock.reset_mock()
        download_url_mock.side_effect = None
        download_url_mock.return_value = 404, None
        try:
            c.get_waveforms("A", "B", "C", "D", UTCDateTime() - 100,
                            UTCDateTime())
        except Exception:
            pass
        assert base_url_ds in download_url_mock.call_args_list[0][0][0]

        # Test the station downloading.
        download_url_mock.reset_mock()
        download_url_mock.side_effect = None
        download_url_mock.return_value = 404, None
        try:
            c.get_stations()
        except Exception:
            pass
        assert base_url_station in download_url_mock.call_args_list[0][0][0]

        # Test the event downloading.
        download_url_mock.reset_mock()
        download_url_mock.side_effect = None
        download_url_mock.return_value = 404, None
        try:
            c.get_events()
        except Exception:
            pass
        assert base_url_event in download_url_mock.call_args_list[0][0][0]

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_no_data_exception(self, download_url_mock):
        """
        Verify that a request returning no data raises an identifiable
        exception
        """
        download_url_mock.return_value = (204, None)
        with pytest.raises(FDSNNoDataException):
            self.client.get_stations()

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_request_too_large_exception(self, download_url_mock):
        """
        Verify that a request returning too much data raises an identifiable
        exception
        """
        download_url_mock.return_value = (413, None)
        with pytest.raises(FDSNRequestTooLargeException):
            self.client.get_stations()

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_timeout_exception(self, download_url_mock):
        """
        Verify that a request timing out raises an identifiable exception
        """
        download_url_mock.return_value = (None, "timeout")
        with pytest.raises(FDSNTimeoutException):
            self.client.get_stations()

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_not_implemented_exception(self, download_url_mock):
        """
        Verify that a client receiving a 501 'Not Implemented' status
        raises an identifiable exception
        """
        download_url_mock.return_value = (501, None)
        with pytest.raises(FDSNNotImplementedException):
            self.client.get_stations()

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_bad_gateway_exception(self, download_url_mock):
        """
        Verify that a client receiving a 502 'Bad Gateway' status
        raises an identifiable exception
        """
        download_url_mock.return_value = (502, None)
        with pytest.raises(FDSNBadGatewayException):
            self.client.get_stations()

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_service_unavailable_exception(self, download_url_mock):
        """
        Verify that opening a client to a service temporarily unavailable
        raises an identifiable exception
        """
        download_url_mock.return_value = (503, None)
        with pytest.raises(FDSNServiceUnavailableException):
            self.client.get_stations()

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_bad_request_exception(self, download_url_mock):
        """
        Verify that a bad request raises an identifiable exception
        """
        download_url_mock.return_value = (400, io.BytesIO(b""))
        with pytest.raises(FDSNBadRequestException):
            self.client.get_stations()

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_server_exception(self, download_url_mock):
        """
        Verify that a server error raises an identifiable exception
        """
        download_url_mock.return_value = (500, None)
        with pytest.raises(FDSNInternalServerException):
            self.client.get_stations()

    @mock.patch("obspy.clients.fdsn.client.download_url")
    def test_too_many_requests_exception(self, download_url_mock):
        """
        Verify that too many requests raise an identifiable exception
        """
        download_url_mock.return_value = (429, None)
        with pytest.raises(FDSNTooManyRequestsException):
            self.client.get_stations()

    def test_iris_earthscope_message(self):
        """
        Test that using "IRIS" short URL in FDSN client shows a warning message
        and switches to "EARTHSCOPE" short URL.
        """
        msg = ("IRIS is now EarthScope, please consider changing the FDSN "
               "client short URL to 'EARTHSCOPE'.")
        with CatchAndAssertWarnings(expected=[(ObsPyDeprecationWarning, msg)]):
            client = Client('IRIS', _discover_services=False)
        assert client.base_url == 'https://service.earthscope.org'

    def test_resif_deprecation_message(self):
        """
        Test that using "RESIF" short URL in FDSN client shows a warning
        message.
        """
        msg = ("RESIF is now EPOSFR. Webservices and client will be "
               "shutdown in 2026. Please consider changing the FDSN "
               "client short URL to 'EPOSFR'.")
        with CatchAndAssertWarnings(expected=[(ObsPyDeprecationWarning, msg)]):
            Client('RESIF', _discover_services=False)

    def test_query_a_non_existent_service_exception(self):
        """
        Tests that a FDSNNoServiceException is raised when no services are
        available but a get_(waveforms|stations|events) method is called
        nonetheless
        """
        start = UTCDateTime(2020, 1, 1)
        end = start + 10

        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        _discover_services=False)
        client.services.pop('dataselect')
        with pytest.raises(FDSNNoServiceException):
            client.get_waveforms('G', 'PEL', '*', 'LHZ', start, end)
        with pytest.raises(FDSNNoServiceException):
            client.get_waveforms_bulk('G', 'PEL', '*', 'LHZ', start, end)

        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        _discover_services=False)
        client.services.pop('station')
        with pytest.raises(FDSNNoServiceException):
            client.get_stations('G', 'PEL', '*', 'LHZ', start, end)
        with pytest.raises(FDSNNoServiceException):
            client.get_stations_bulk('G', 'PEL', '*', 'LHZ', start, end)

        client = Client(base_url="EARTHSCOPE", user_agent=USER_AGENT,
                        _discover_services=False)
        client.services.pop('event')
        with pytest.raises(FDSNNoServiceException):
            client.get_events(start, end)

    @mock.patch('obspy.clients.fdsn.client.download_url')
    def test_no_gzip(self, download_mock):
        """
        Tests that opting out of gzip works
        """
        for use_gzip in (True, False):
            client = Client(_discover_services=False, use_gzip=use_gzip)
            # check a station request
            bio = io.BytesIO()
            Inventory().write(bio, 'STATIONXML')
            download_mock.return_value = (200, bio)
            client.get_stations(network='XX')
            # check that the download url was called with the correct option
            # for use_gzip
            assert (
                download_mock.call_args.kwargs.get('use_gzip', None)
                is use_gzip)
            # check an event request
            bio = io.BytesIO()
            Catalog().write(bio, 'QUAKEML')
            download_mock.return_value = (200, bio)
            client.get_events(minmagnitude=10)
            # check that the download url was called with the correct option
            # for use_gzip
            assert (
                download_mock.call_args.kwargs.get('use_gzip', None)
                is use_gzip)
            # check dataselct request, these always have gzip disabled
            bio = io.BytesIO()
            read().write(bio, 'MSEED')
            download_mock.return_value = (200, bio)
            now = UTCDateTime()
            client.get_waveforms("XX", "XXX", "", "XXX", now - 10, now - 5)
            # check that the download url was called with the correct option
            # for use_gzip
            assert (
                download_mock.call_args.kwargs.get('use_gzip', None)
                is False)

    def test_raise_on_error(self):
        """
        Test helper function raise on error to prevent more regression bugs
        like introduced inadvertently with #3306 which took some while to get
        properly fixed with multiple cases that ended up raising unexpected
        exceptions and not proper FDSNException like expected by e.g. Mass
        Downloader
        """
        server_string = 'detailed server response here.'
        server_bytes = b'detailed server response here.'

        # a) test a proper HTTP response with an error which we handle
        #    explicitly
        code = 204
        msg = (f'No data available for request.\nHTTP Status code: '
               f'{code}\nDetailed response of server:\n\n{server_string}')
        data = io.BytesIO(server_bytes)
        with pytest.raises(FDSNNoDataException, match=msg):
            raise_on_error(code, data)
        # sometimes it seems the server response comes as a StringIO for
        # whatever reason, so we took care to handle that too
        data = io.StringIO(server_string)
        with pytest.raises(FDSNNoDataException, match=msg):
            raise_on_error(code, data)

        # b) test a proper HTTP response with a HTTP code we don't handle
        #    specifically
        code = 999
        msg = (f'Unknown HTTP code: {code}\nDetailed response of server:'
               f'\n\n{server_string}')
        data = io.BytesIO(server_bytes)
        with pytest.raises(FDSNException, match=msg):
            raise_on_error(code, data)
        # sometimes it seems the server response comes as a StringIO for
        # whatever reason, so we took care to handle that too
        data = io.StringIO(server_string)
        with pytest.raises(FDSNException, match=msg):
            raise_on_error(code, data)

        # c) test some cases that encounter exceptions ahead of proper HTTP
        #    responses, like random network failures like timeout errors or
        #    address resolution errors.

        #    This case is reconstructed from bug report #3496
        code = None
        data = URLError(
            socket.gaierror(-3, 'Temporary failure in name resolution'))
        msg = (r'Unknown Error \(URLError\):.*Temporary failure in name '
               r'resolution')
        with pytest.raises(FDSNException, match=msg):
            raise_on_error(code, data)
        #    This case is reconstructed from bug report #3510. Python <=3.9
        #    this was encountering socket.timeout object, which since 3.10 is
        #    just an alias for TimeoutError, so it depends on Python version
        #    what we actually encounter here
        code = None
        try:
            data = socket.timeout()
        except AttributeError:
            # if the deprecated alias socket.timeout for TimeoutError ever gets
            # removed we can be ready for it here
            data = TimeoutError()
        msg = 'Timed Out'
        with pytest.raises(FDSNException, match=msg):
            raise_on_error(code, data)



# -*- coding: utf-8 -*-
MODULE_NAME = "obspy.io.segy"


def _patch_header(header, ebcdic=False):
    """
    Helper function to patch a textual header to include the revision
    number and the end header mark.
    """
    revnum = "C39 SEG Y REV1"

    if ebcdic:
        revnum = revnum.encode("EBCDIC-CP-BE")
        end_header = "C40 END EBCDIC        ".encode("EBCDIC-CP-BE")
    else:
        revnum = revnum.encode("ascii")
        end_header = "C40 END TEXTUAL HEADER".encode("ascii")

    header = header[:3200-160] + revnum + header[3200-146:]
    header = header[:3200-80] + end_header + header[3200-58:]

    return header
